"""
This module wraps the parse function of the Element Tree library to parse XML files with a
namespace map. Element tree discards all prefixes when parsing the file.
It is used by the different parsing modules.
"""

import xml.etree.ElementTree as ET
from io import StringIO
from weakref import WeakKeyDictionary

# Global storage for namespace maps, keyed by Element objects
# Using WeakKeyDictionary so entries are automatically removed when elements are garbage collected
_ns_map_store: WeakKeyDictionary[ET.Element, dict[str, str]] = WeakKeyDictionary()


def get_ns_map(elem: ET.Element) -> dict[str, str]:
    """
    Get the namespace map for an element.
    :param elem: The XML element
    :return: The namespace prefix to URI mapping dict
    """
    return _ns_map_store.get(elem, {})


def set_ns_map(elem: ET.Element, ns_map: dict[str, str]) -> None:
    """
    Set the namespace map for an element.
    :param elem: The XML element
    :param ns_map: The namespace prefix to URI mapping dict
    """
    _ns_map_store[elem] = ns_map


def parse_file(file: str | StringIO) -> ET.ElementTree:
    """
    Parses a file, returns the Root element with namespace maps stored for each element.
    Use get_ns_map(element) to retrieve the namespace map for any element.
    :param file: either the file path (str) or a file-like object
    :return: The parsed ElementTree
    """
    events = "start", "start-ns", "end-ns"

    root = None
    ns_map: list[tuple[str, str]] = []

    for event, elem in ET.iterparse(file, events):
        if event == "start-ns":
            # elem is a tuple[str, str] when event is "start-ns"
            ns_map.append(elem)  # type: ignore[arg-type]
        elif event == "end-ns":
            ns_map.pop()
        elif event == "start":
            if root is None:
                root = elem
            set_ns_map(elem, dict(ns_map))

    return ET.ElementTree(root)



"""
This module contains all classes and functions necessary for parsing Taxonomy schema files.
"""

import json
import logging
import os
import xml.etree.ElementTree as ET
from collections import OrderedDict
from urllib.parse import unquote

from xbrl import TaxonomyNotFound
from xbrl.cache import HttpCache
from xbrl.helper.uri_helper import compare_uri, is_url, resolve_uri
from xbrl.linkbase import (
    ExtendedLink,
    Label,
    LabelArc,
    Linkbase,
    LinkbaseType,
    parse_linkbase,
    parse_linkbase_url,
)
from xbrl.ns_map import NS_MAP

logger = logging.getLogger(__name__)

LINK_NS: str = "{http://www.xbrl.org/2003/linkbase}"
XLINK_NS: str = "{http://www.w3.org/1999/xlink}"
XDS_NS: str = "{http://www.w3.org/2001/XMLSchema}"
XBRLI_NS: str = "{http://www.xbrl.org/2003/instance}"

# dictionary containing all common prefixes and the corresponding namespaces.
NAME_SPACES: dict = {
    "xsd": "http://www.w3.org/2001/XMLSchema",
    "link": "http://www.xbrl.org/2003/linkbase",
    "xlink": "http://www.w3.org/1999/xlink",
    "xbrldt": "http://xbrl.org/2005/xbrldt",
}


class Concept:
    """
    Class representing a Concept defined in the schema (xs:element)
    i.e:
    <xs:element id='us-gaap_Assets' name='Assets' nillable='true'
    substitutionGroup='xbrli:item' type='xbrli:monetaryItemType'
    xbrli:balance='debit' xbrli:periodType='instant' />
    """

    def __init__(self, xml_id: str, schema_url: str, name: str) -> None:
        """
        :param xml_id: Id of the concept in the xml
        :param schema_url: url of the schema in which the concept is defined
        :param name: name of the concept
        """
        self.xml_id: str = xml_id
        self.schema_url: str = schema_url
        self.name: str = name
        self.substitution_group: str | None = None
        self.concept_type: str | None = None
        self.abstract: bool | None = None
        self.nillable: bool | None = None
        self.period_type: str | None = None
        self.balance: str | None = None
        self.labels: list[Label] = []

    def to_dict(self):
        """
        Converts the Concept object into a dictionary representation
        """
        return {
            "xml_id": self.xml_id,
            "schema_url": self.schema_url,
            "name": self.name,
            "substitution_group": self.substitution_group,
            "concept_type": self.concept_type,
            "abstract": self.abstract,
            "nillable": self.nillable,
            "period_type": self.period_type,
            "balance": self.balance,
            # Assuming Label class has to_dict()
            "labels": [label.to_dict() for label in self.labels] if self.labels else [],
        }

    def to_json(self):
        """
        Converts the Concept object into a JSON string
        """
        return json.dumps(self.to_dict(), indent=4)

    def __str__(self) -> str:
        return self.name


class ExtendedLinkRole:
    """
    Class representing a ELR.
    A ELR is a set of relations representing a piece of the report (i.e. "1003000 - Statement - Consolidated Balance Sheets")
    ELR's a used to separate Relation linkbases into smaller logical chunks, so it is commonly referenced in the
    calculation, definition and presentation linkbases
    """

    def __init__(self, role_id: str, uri: str, definition: str) -> None:
        """

        :param role_id:
        :param uri:
        :param definition:
        """
        self.xml_id: str = role_id
        self.uri: str = uri
        self.definition: str = definition
        self.definition_link: ExtendedLink | None = None
        self.presentation_link: ExtendedLink | None = None
        self.calculation_link: ExtendedLink | None = None

    def __str__(self) -> str:
        return self.definition


class TaxonomySchema:
    """
    Class represents a Generic Taxonomy Schema. Since this parser is optimized for EDGAR submission's,
    it will only differentiate between the Extending Taxonomy (the taxonomy that comes with the filing) and
    multiple base Taxonomies (i.e dei, us-gaap, exch, naics, sic ...).
    This parser will not parse all Schemas and imports, only what is necessary.
    """

    def __init__(self, schema_url: str, namespace: str):
        """
        The imports array stores an array of all Schemas that are imported.


        :param schema_url:
        :param namespace:
        """
        self.imports: list[TaxonomySchema] = []
        self.link_roles: list[ExtendedLinkRole] = []
        self.lab_linkbases: list[Linkbase] = []
        self.def_linkbases: list[Linkbase] = []
        self.cal_linkbases: list[Linkbase] = []
        self.pre_linkbases: list[Linkbase] = []

        self.schema_url = schema_url
        self.namespace = namespace
        # store the concepts in a dictionary with the concept_id as key
        self.concepts: dict = {}
        # The linkbases reference concepts by their id, the instance file by name.
        # In order to get O(1) in both cases, create a dictionary where the id of a concept can be looked up,
        # based on the name
        self.name_id_map: dict = {}

    def __str__(self) -> str:
        return self.namespace

    def get_taxonomy(self, url: str, visited: None | set[str] = None):
        """
        Returns the taxonomy with the given namespace (if it is the current taxonomy, or if it is imported)
        If the taxonomy cannot be found, the function will return None
        :param url: can either be the namespace or the schema url
        :param visited: set of already visited schema urls to prevent infinite recursion
        :return: either a TaxonomySchema obj or None
        """
        if visited is None:
            visited = set()

        # Prevent infinite recursion from circular imports
        if self.schema_url in visited:
            return None
        visited.add(self.schema_url)

        if compare_uri(self.namespace, url) or compare_uri(self.schema_url, url):
            return self

        for imported_tax in self.imports:
            result = imported_tax.get_taxonomy(url, visited)
            if result is not None:
                return result
        return None

    def get_schema_urls(self, visited: None | set[str] = None) -> list[str]:
        """
        Returns an array of all taxonomy urls that are used by this taxonomy
        Also includes the schema url of this taxonomy
        :return:
        """
        if visited is None:
            visited = set()
        # IF the taxonomy imports have already been added to the list, do not add them again
        if self.schema_url in visited:
            return []
        visited.add(self.schema_url)

        urls: list[str] = [self.schema_url]
        for imported_tax in self.imports:
            urls += imported_tax.get_schema_urls(visited)
        return list(set(urls))


class TaxonomyParser:
    """
    Helper class to parse taxonomies and cache namespace maps
    """

    """
    :param use_local_ns_map: if enabled the parser will use a local namespace map as fallback to try resolving taxonomies
    :param fetch_edgar_taxonomies: if enabled, the parser will upfront load the EDGAR Common Taxonomies from
        https://www.sec.gov/files/edgartaxonomies.xml and use them as fallback
    """

    def __init__(
        self,
        cache: HttpCache,
        use_local_ns_map: bool = True,
        fetch_edgar_taxonomies: bool = False,
        fetch_py_xbrl_ns_map: bool = True,
        max_taxonomy_cache_size: int = 60,
    ) -> None:
        self.cache = cache
        self.use_local_ns_map = use_local_ns_map
        self.fetch_edgar_taxonomies = fetch_edgar_taxonomies
        self.fetch_py_xbrl_ns_map = fetch_py_xbrl_ns_map
        self.max_taxonomy_cache_size = max_taxonomy_cache_size

        # Cache for global namespace to schema url mapping of common taxonomies
        self.global_ns_map: dict[str, str] = {}
        # Cache for parsed taxonomies with LRU eviction, the key is the schema url
        self.taxonomy_cache: OrderedDict[str, TaxonomySchema] = OrderedDict()

        if self.use_local_ns_map:
            self._add_local_ns_map()
        if self.fetch_edgar_taxonomies:
            self._add_edgar_taxonomies()

    def _add_local_ns_map(self):
        """
        Adds the local NS_MAP to the global namespace map
        """
        for ns, url in NS_MAP.items():
            self.global_ns_map[ns] = url

    def _add_edgar_taxonomies(self):
        """
        Adds the Edgar taxonomy namespace map to the global namespace map
        """
        edgar_taxonomies_url = "https://www.sec.gov/files/edgartaxonomies.xml"
        edgar_taxonomies_path = self.cache.cache_file(edgar_taxonomies_url)
        root: ET.Element = ET.parse(edgar_taxonomies_path).getroot()

        for loc in root.findall("Loc"):
            namespace_el = loc.find("Namespace")
            href_el = loc.find("Href")

            if namespace_el is not None and href_el is not None:
                namespace = namespace_el.text.strip()
                href = href_el.text.strip()
                self.global_ns_map[namespace] = href

    def _add_to_cache(self, schema_path: str, taxonomy: TaxonomySchema):
        self.taxonomy_cache[schema_path] = taxonomy
        self.taxonomy_cache.move_to_end(schema_path)

        # Remove oldest entry if cache exceeds max size
        if len(self.taxonomy_cache) > self.max_taxonomy_cache_size:
            self.taxonomy_cache.popitem(last=False)

    def _is_in_cache(self, schema_path: str) -> bool:
        return schema_path in self.taxonomy_cache

    def _load_from_cache(self, schema_path: str) -> TaxonomySchema | None:
        if self._is_in_cache(schema_path):
            # Move to end (mark as recently used)
            self.taxonomy_cache.move_to_end(schema_path)
            return self.taxonomy_cache[schema_path]
        return None

    def try_taxonomy_from_namespace(self, namespace: str) -> TaxonomySchema:
        """
        Tries to parse a taxonomy by its namespace using the global namespace map

        :param namespace: Namespace of the taxonomy
        :return: Parsed TaxonomySchema object or None if not found
        """
        if namespace in self.global_ns_map:
            schema_url = self.global_ns_map[namespace]
            return self.parse_taxonomy(schema_url)
        raise TaxonomyNotFound(namespace)

    def parse_taxonomy(
        self,
        schema_path: str,
        imported_schema_uris: set = set(),
        schema_url: str | None = None,
    ) -> TaxonomySchema:
        """
        Parses a taxonomy schema file.

        :param schema_path: url to the schema (on the internet)
        :param cache: :class:`xbrl.cache.HttpCache` instance
        :param imported_schema_uris: set of already imported schema uris
        :param schema_url: if this url is set, the script will try to fetch additionally imported files such as linkbases or
            imported schemas from the remote location. If this url is None, the script will try to find those resources locally.
        :return: parsed :class:`xbrl.taxonomy.TaxonomySchema` object
        """
        schema_path = str(schema_path).strip()
        if is_url(schema_path):
            # If the path is acually a url, we set the schema_url to it and download the file first
            schema_url = schema_path
            schema_path = self.cache.cache_file(schema_path)

        if not os.path.exists(schema_path):
            raise TaxonomyNotFound(f"Could not find taxonomy schema at {schema_path}")

        cached = self._load_from_cache(schema_path)
        if cached is not None:
            return cached

        # Get the local absolute path to the schema file (and download it if it is not yet cached)
        root: ET.Element = ET.parse(schema_path).getroot()
        # get the target namespace of the taxonomy
        target_ns = root.attrib["targetNamespace"]
        taxonomy: TaxonomySchema = TaxonomySchema(
            schema_url if schema_url else schema_path, target_ns
        )

        import_elements: list[ET.Element] = root.findall("xsd:import", NAME_SPACES)

        for import_element in import_elements:
            import_uri = import_element.attrib["schemaLocation"].strip()

            # Skip empty imports
            if import_uri == "":
                continue

            # Skip already imported URIs
            if import_uri in imported_schema_uris:
                continue

            # sometimes the import schema location is relative. i.e schemaLocation="xbrl-linkbase-2003-12-31.xsd"
            if is_url(import_uri):
                # fetch the schema file from remote
                taxonomy.imports.append(self.parse_taxonomy(import_uri))
            elif schema_url:
                # fetch the schema file from remote by reconstructing the full url
                import_url = resolve_uri(schema_url, import_uri)
                imported_schema_uris.add(import_uri)
                taxonomy.imports.append(self.parse_taxonomy(import_url))
            else:
                # We have to try to fetch the linkbase locally because no full url can be constructed
                import_path = resolve_uri(schema_path, import_uri)
                taxonomy.imports.append(
                    self.parse_taxonomy(import_path, imported_schema_uris)
                )

        role_type_elements: list[ET.Element] = root.findall(
            "xsd:annotation/xsd:appinfo/link:roleType", NAME_SPACES
        )
        # parse ELR's
        for elr_elem in role_type_elements:
            elr_definition = elr_elem.find(LINK_NS + "definition")
            if elr_definition is None or elr_definition.text is None:
                continue
            taxonomy.link_roles.append(
                ExtendedLinkRole(
                    elr_elem.attrib["id"],
                    elr_elem.attrib["roleURI"],
                    elr_definition.text.strip(),
                )
            )

        # find all elements that are defined in the schema
        for element in root.findall(XDS_NS + "element"):
            # if a concept has no id, it can not be referenced by a linkbase, so just ignore it
            if "id" not in element.attrib or "name" not in element.attrib:
                continue
            el_id: str = element.attrib["id"]
            el_name: str = element.attrib["name"]

            new_concept = Concept(el_id, schema_url or schema_path, el_name)
            new_concept.concept_type = (
                element.attrib["type"] if "type" in element.attrib else None
            )
            new_concept.nillable = (
                bool(element.attrib["nillable"])
                if "nillable" in element.attrib
                else False
            )
            new_concept.abstract = (
                bool(element.attrib["abstract"])
                if "abstract" in element.attrib
                else False
            )
            type_attr_name = XBRLI_NS + "periodType"
            new_concept.period_type = (
                element.attrib[type_attr_name]
                if type_attr_name in element.attrib
                else None
            )
            balance_attr_name = XBRLI_NS + "balance"
            new_concept.balance = (
                element.attrib[balance_attr_name]
                if balance_attr_name in element.attrib
                else None
            )
            # remove the prefix from the substitutionGroup (i.e xbrli:item -> item)
            new_concept.substitution_group = (
                element.attrib["substitutionGroup"].split(":")[-1]
                if "substitutionGroup" in element.attrib
                else None
            )

            taxonomy.concepts[new_concept.xml_id] = new_concept
            taxonomy.name_id_map[new_concept.name] = new_concept.xml_id

        linkbase_ref_elements: list[ET.Element] = root.findall(
            "xsd:annotation/xsd:appinfo/link:linkbaseRef", NAME_SPACES
        )
        for linkbase_ref in linkbase_ref_elements:
            linkbase_uri = linkbase_ref.attrib[XLINK_NS + "href"]
            role = (
                linkbase_ref.attrib[XLINK_NS + "role"]
                if XLINK_NS + "role" in linkbase_ref.attrib
                else None
            )
            linkbase_type = (
                LinkbaseType.get_type_from_role(role)
                if role is not None
                else LinkbaseType.guess_linkbase_role(linkbase_uri)
            )
            if linkbase_type is None:
                logger.info(f"Ignoring unsupported linkbase: {role}")
                continue

            # check if the linkbase url is relative
            linkbase: Linkbase
            if is_url(linkbase_uri):
                # fetch the linkbase from remote
                linkbase = parse_linkbase_url(linkbase_uri, linkbase_type, self.cache)
            elif schema_url:
                # fetch the linkbase from remote by reconstructing the full URL
                linkbase_url = resolve_uri(schema_url, linkbase_uri)
                linkbase = parse_linkbase_url(linkbase_url, linkbase_type, self.cache)
            else:
                # We have to try to fetch the linkbase locally because no full url can be constructed
                linkbase_path = resolve_uri(schema_path, linkbase_uri)
                linkbase = parse_linkbase(linkbase_path, linkbase_type)

            # add the linkbase to the taxonomy
            if linkbase_type == LinkbaseType.DEFINITION:
                taxonomy.def_linkbases.append(linkbase)
            elif linkbase_type == LinkbaseType.CALCULATION:
                taxonomy.cal_linkbases.append(linkbase)
            elif linkbase_type == LinkbaseType.PRESENTATION:
                taxonomy.pre_linkbases.append(linkbase)
            elif linkbase_type == LinkbaseType.LABEL:
                taxonomy.lab_linkbases.append(linkbase)

        # loop over the ELR's of the schema and assign the extended links from the linkbases
        all_def_links = [
            def_linkbase.extended_links for def_linkbase in taxonomy.def_linkbases
        ]
        all_pre_links = [
            pre_linkbase.extended_links for pre_linkbase in taxonomy.pre_linkbases
        ]
        all_cal_links = [
            cal_linkbase.extended_links for cal_linkbase in taxonomy.cal_linkbases
        ]
        for elr in taxonomy.link_roles:
            for extended_def_links in all_def_links:
                for extended_def_link in extended_def_links:
                    if extended_def_link.elr_id is None:
                        continue
                    if extended_def_link.elr_id.split("#")[1] == elr.xml_id:
                        elr.definition_link = extended_def_link
                        break
            for extended_pre_links in all_pre_links:
                for extended_pre_link in extended_pre_links:
                    if extended_pre_link.elr_id is None:
                        continue
                    if extended_pre_link.elr_id.split("#")[1] == elr.xml_id:
                        elr.presentation_link = extended_pre_link
                        break
            for extended_cal_links in all_cal_links:
                for extended_cal_link in extended_cal_links:
                    if extended_cal_link.elr_id is None:
                        continue
                    if extended_cal_link.elr_id.split("#")[1] == elr.xml_id:
                        elr.calculation_link = extended_cal_link
                        break

        for label_linkbase in taxonomy.lab_linkbases:
            for extended_link in label_linkbase.extended_links:
                for root_locator in extended_link.root_locators:
                    # find the taxonomy the locator is referring to
                    schema_url, concept_id = unquote(root_locator.href).split("#")
                    c_taxonomy: TaxonomySchema = taxonomy.get_taxonomy(schema_url)
                    if c_taxonomy is None:
                        if schema_url in NS_MAP.values():
                            c_taxonomy = self.parse_taxonomy(schema_url)
                            taxonomy.imports.append(c_taxonomy)
                        else:
                            continue
                    concept: Concept = c_taxonomy.concepts[concept_id]
                    concept.labels = []
                    for label_arc in root_locator.children:
                        if isinstance(label_arc, LabelArc):
                            for label in label_arc.labels:
                                concept.labels.append(label)

        self._add_to_cache(schema_path, taxonomy)
        return taxonomy



"""
This unittest tests the parsing of locally saved instance documents.
"""

import sys
import unittest
from xbrl.cache import HttpCache
from xbrl.instance import parse_ixbrl, parse_xbrl, XbrlInstance
import logging


# abs_file_path: str = str(os.path.dirname(os.path.abspath(inspect.getframeinfo(inspect.currentframe()).filename)))


class InstanceTest(unittest.TestCase):
    """
    Unit test for taxonomy.test_parse_taxonomy()
    """

    def test_parse_xbrl_document(self):
        """Integration test for instance.parse_xbrl_instance()"""
        logging.basicConfig(stream=sys.stdout, level=logging.INFO)
        cache_dir: str = "./cache/"
        cache: HttpCache = HttpCache(cache_dir)

        instance_doc_url: str = "./tests/data/example.xml"
        inst: XbrlInstance = parse_xbrl(instance_doc_url, cache)
        print(inst)
        self.assertEqual(len(inst.facts), 1)

    def test_parse_ixbrl_document(self):
        """Integration test for instance.parse_ixbrl_instance()"""
        logging.basicConfig(stream=sys.stdout, level=logging.INFO)
        cache_dir: str = "./cache/"
        cache: HttpCache = HttpCache(cache_dir)

        instance_doc_url: str = "./tests/data/example.html"
        inst: XbrlInstance = parse_ixbrl(instance_doc_url, cache)
        print(inst)
        self.assertEqual(len(inst.facts), 3)


if __name__ == "__main__":
    unittest.main()



# This library is a simple implementation of a function to convert textual
# numbers written in English into their integer representations.
#
# This code is open source according to the MIT License as follows.
#
# Copyright (c) 2008 Greg Hewgill
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
# https://github.com/ghewgill/text2num/

import re

Small = {
    "zero": 0,
    "one": 1,
    "two": 2,
    "three": 3,
    "four": 4,
    "five": 5,
    "six": 6,
    "seven": 7,
    "eight": 8,
    "nine": 9,
    "ten": 10,
    "eleven": 11,
    "twelve": 12,
    "thirteen": 13,
    "fourteen": 14,
    "fifteen": 15,
    "sixteen": 16,
    "seventeen": 17,
    "eighteen": 18,
    "nineteen": 19,
    "twenty": 20,
    "thirty": 30,
    "forty": 40,
    "fifty": 50,
    "sixty": 60,
    "seventy": 70,
    "eighty": 80,
    "ninety": 90,
}

Magnitude = {
    "thousand": 1000,
    "million": 1000000,
    "billion": 1000000000,
    "trillion": 1000000000000,
    "quadrillion": 1000000000000000,
    "quintillion": 1000000000000000000,
    "sextillion": 1000000000000000000000,
    "septillion": 1000000000000000000000000,
    "octillion": 1000000000000000000000000000,
    "nonillion": 1000000000000000000000000000000,
    "decillion": 1000000000000000000000000000000000,
}


class NumberException(Exception):
    def __init__(self, msg):
        Exception.__init__(self, msg)


def text2num(s) -> int:
    if s.isnumeric():
        return int(s)

    a = re.split(r"[\s-]+", s)
    n = 0
    g = 0
    for w in a:
        x = Small.get(w, None)
        if x is not None:
            g += x
        elif w == "hundred" and g != 0:
            g *= 100
        else:
            x = Magnitude.get(w, None)
            if x is not None:
                n += g * x
                g = 0
            else:
                raise NumberException("Unknown number: " + w)
    return n + g



"""
This is just a very simple unit test, that checks if the HttpCache is able to download a file, save it in the
specified cache directory and serves it from there.
"""

import logging
import os
import sys
import tempfile
import time
import unittest

from xbrl.cache import HttpCache


class CacheHelperTest(unittest.TestCase):
    def test_cache_file(self):
        """
        Unit test for CacheHelper.cache_file
        :return:
        """
        logging.basicConfig(stream=sys.stdout, level=logging.INFO)
        with tempfile.TemporaryDirectory(prefix="py-xbrl-cache-test-") as tmp_cache_dir:
            cache_dir: str = os.path.join(tmp_cache_dir, "")
            delay: int = 5000
            cache: HttpCache = HttpCache(cache_dir, delay)

            test_url: str = "https://www.w3schools.com/xml/note.xml"
            expected_path: str = os.path.join(cache_dir, "www.w3schools.com", "xml", "note.xml")

            # if the testing file already exists delete if first
            if os.path.isfile(expected_path):
                os.remove(expected_path)

            # on the first execution the file will be downloaded from the internet, no delay for first download
            time_stamp: float = time.time()
            self.assertEqual(cache.cache_file(test_url), expected_path)
            time_delta = time.time() - time_stamp
            self.assertLess(time_delta, delay / 1000)
            logging.info(f"Time delta for first download: {time_delta}ms")

            # delete the file and download it again to check if the delay for the second download is working
            self.assertTrue(cache.purge_file(test_url))

            time_stamp: float = time.time()
            self.assertEqual(cache.cache_file(test_url), expected_path)
            time_delta = time.time() - time_stamp
            self.assertGreaterEqual(time_delta, delay / 1000)
            logging.info(f"Time delta for second download: {time_delta}ms")

            # now that the file is cached on the hard drive, the file path should be returned immediately
            time_stamp = time.time()
            self.assertEqual(cache.cache_file(test_url), expected_path)
            time_delta = time.time() - time_stamp
            self.assertLess(time_delta, delay / 1000)
            logging.info(f"Time delta for third download: {time_delta}ms")

            # test if the file was downloaded
            self.assertTrue(os.path.isfile(expected_path))
            # delete the file
            self.assertTrue(cache.purge_file(test_url))
            # test if the file was deleted
            self.assertFalse(os.path.isfile(expected_path))


if __name__ == "__main__":
    unittest.main()



"""
Module containing functions for creating and resolving uri's
"""

import os
import re


def resolve_uri(dir_uri: str, relative_uri: str) -> str:
    """
    Returns a complete absolute uri.
    i.e
    dir_uri is 'http://abc.org/a/b/c'
    relative_uri: '/../lab.xml'
        the function would resolve the absolute uri to: http://abc.org/a/c/lab.xml

    if the relative_uri is already absolute, the function will just return the relative_url
    @param dir_uri:
    @param relative_uri:
    @return:
    """
    if is_url(relative_uri):
        return relative_uri

    # remove redundant characters in the relative uri
    if relative_uri.startswith("/"):
        relative_uri = relative_uri[1:]
    if relative_uri.startswith("./"):
        relative_uri = relative_uri[2:]

    dir_uri = str(dir_uri)
    if not is_url(dir_uri):
        # check if the dir_uri was really a path to a directory or a file
        if "." in dir_uri.split(os.sep)[-1]:
            return os.path.normpath(os.path.dirname(dir_uri) + os.sep + relative_uri)
        else:
            return os.path.normpath(dir_uri + os.sep + relative_uri)

    # === From here on we only process urls ===
    # remove the file name if the dir_uri was the url to a file
    if "." in dir_uri.split("/")[-1]:
        dir_uri = "/".join(dir_uri.split("/")[0:-1])
    if not dir_uri.endswith("/"):
        dir_uri += "/"

    absolute_uri = dir_uri + relative_uri
    if not is_url(dir_uri):
        # make sure the path is correct
        absolute_uri = os.path.normpath(absolute_uri)

    url_parts = absolute_uri.split("/")
    for x in range(0, absolute_uri.count("/..")):
        # loop over the url_parts array and remove the path_part, that has a '..' after it
        for y in range(0, len(url_parts) - 1):
            if url_parts[y + 1] == "..":
                del url_parts[y]  # delete the path part affected by the '/../'
                del url_parts[y]  # delete the '/../' itself
                break

    return "/".join(url_parts)


def compare_uri(uri1: str, uri2: str) -> bool:
    """
    Compares two uri's and returns true if they are considered equal and false if not.
    Examples:
        - uri1: http://abc.de/2020, uri2: https://abc.de/2020 -> True
        - uri1: ./abc.de/2020, uri2: abc.de\\2020 -> True
        - uri1: /abc.de/2020, uri2: /abc/2020 -> False
    :param uri1:
    :param uri2:
    :return:
    """
    # first remove any protocol
    if "://" in uri1:
        uri1 = uri1.split("://")[1]
    if "://" in uri2:
        uri2 = uri2.split("://")[1]

    uri1_segments: list[str] = re.findall(r"[\w']+", uri1)
    uri2_segments: list[str] = re.findall(r"[\w']+", uri2)
    return uri1_segments == uri2_segments


def is_url(candidate: str) -> bool:
    return candidate.lower().startswith("http://") or candidate.lower().startswith(
        "https://"
    )



"""
This unittest tests the parsing of locally saved taxonomies
"""

import sys
import unittest
from xbrl.cache import HttpCache
from xbrl.taxonomy import  TaxonomySchema, TaxonomyParser
import logging


class TaxonomySchemaTest(unittest.TestCase):
    """
    Unit test for taxonomy.test_parse_taxonomy()
    """

    def test_parse_taxonomy(self):
        logging.basicConfig(stream=sys.stdout, level=logging.INFO)
        cache_dir: str = "./cache/"
        cache: HttpCache = HttpCache(cache_dir)
        print(f"Saving to {cache_dir}")

        parser = TaxonomyParser(cache)

        extension_schema_path: str = "./tests/data/example.xsd"
        # extension_schema_path: str = './data/example.xsd'
        tax: TaxonomySchema = parser.parse_taxonomy(extension_schema_path)
        print(tax)
        srt_tax: TaxonomySchema = tax.get_taxonomy("http://fasb.org/srt/2020-01-31")
        self.assertTrue(srt_tax)
        self.assertEqual(len(srt_tax.concepts), 489)

        # check if the labels where successfully linked to the concept
        self.assertEqual(len(tax.concepts["example_Assets"].labels), 2)


if __name__ == "__main__":
    unittest.main()



"""
This unit test tests the uri resolver.
It is often the case, that a taxonomy schema imports another taxonomy using a relative path.
i.e:
<link:linkbaseRef [..] xlink:href="./../example_lab.xml" [..]/>
The job of the uri resolver is to resolve those relative paths and urls and return an absolute path or url
"""

import logging
import sys
import unittest
import os
from xbrl.helper.uri_helper import resolve_uri, compare_uri

logging.basicConfig(stream=sys.stdout, level=logging.INFO)


class UriResolverTest(unittest.TestCase):
    def test_resolve_uri(self):
        """
        :return:
        """

        test_arr = [
            # test paths
            (
                (
                    "E:\\Programming\\python\\xbrl_parser\\tests\\data\\example.xsd",
                    "/example-lab.xml",
                ),
                os.sep.join(
                    [
                        "E:",
                        "Programming",
                        "python",
                        "xbrl_parser",
                        "tests",
                        "data",
                        "example-lab.xml",
                    ]
                ),
            ),
            (
                (
                    r"E:\Programming\python\xbrl_parser\tests\data\example.xsd",
                    "/example-lab.xml",
                ),
                os.sep.join(
                    [
                        "E:",
                        "Programming",
                        "python",
                        "xbrl_parser",
                        "tests",
                        "data",
                        "example-lab.xml",
                    ]
                ),
            ),
            (
                (
                    "E:/Programming/python/xbrl_parser/tests/data/example.xsd",
                    "/example-lab.xml",
                ),
                os.sep.join(
                    [
                        "E:",
                        "Programming",
                        "python",
                        "xbrl_parser",
                        "tests",
                        "data",
                        "example-lab.xml",
                    ]
                ),
            ),
            # test different path separators
            (
                (
                    "E:\\Programming\\python\\xbrl_parser\\tests\\data/example.xsd",
                    "/example-lab.xml",
                ),
                os.sep.join(
                    [
                        "E:",
                        "Programming",
                        "python",
                        "xbrl_parser",
                        "tests",
                        "data",
                        "example-lab.xml",
                    ]
                ),
            ),
            # test directory traversal
            (
                (
                    "E:/Programming/python/xbrl_parser/tests/data/",
                    "/../example-lab.xml",
                ),
                os.sep.join(
                    [
                        "E:",
                        "Programming",
                        "python",
                        "xbrl_parser",
                        "tests",
                        "example-lab.xml",
                    ]
                ),
            ),
            (
                (
                    "E:/Programming/python/xbrl_parser/tests/data",
                    "./../example-lab.xml",
                ),
                os.sep.join(
                    [
                        "E:",
                        "Programming",
                        "python",
                        "xbrl_parser",
                        "tests",
                        "example-lab.xml",
                    ]
                ),
            ),
            (
                (
                    "E:/Programming/python/xbrl_parser/tests/data/example.xsd",
                    "../../example-lab.xml",
                ),
                os.sep.join(
                    ["E:", "Programming", "python", "xbrl_parser", "example-lab.xml"]
                ),
            ),
            # test urls
            (
                ("http://example.com/a/b/c/d/e/f/g", "file.xml"),
                "http://example.com/a/b/c/d/e/f/g/file.xml",
            ),
            (
                ("http://example.com/a/b/c/d/e/f/g", "/file.xml"),
                "http://example.com/a/b/c/d/e/f/g/file.xml",
            ),
            (
                ("http://example.com/a/b/c/d/e/f/g", "./file.xml"),
                "http://example.com/a/b/c/d/e/f/g/file.xml",
            ),
            (
                ("http://example.com/a/b/c/d/e/f/g", "../file.xml"),
                "http://example.com/a/b/c/d/e/f/file.xml",
            ),
            (
                ("http://example.com/a/b/c/d/e/f/g", "/../file.xml"),
                "http://example.com/a/b/c/d/e/f/file.xml",
            ),
            (
                ("http://example.com/a/b/c/d/e/f/g", "./../file.xml"),
                "http://example.com/a/b/c/d/e/f/file.xml",
            ),
            (
                ("http://example.com/a/b/c/d/e/f/g", "../../file.xml"),
                "http://example.com/a/b/c/d/e/file.xml",
            ),
            (
                ("http://example.com/a/b/c/d/e/f/g", "/../../file.xml"),
                "http://example.com/a/b/c/d/e/file.xml",
            ),
            (
                ("http://example.com/a/b/c/d/e/f/g", "./../../file.xml"),
                "http://example.com/a/b/c/d/e/file.xml",
            ),
            (
                ("http://example.com/a/b/c/d/e/f/g/", "../../../file.xml"),
                "http://example.com/a/b/c/d/file.xml",
            ),
            (
                ("http://example.com/a/b/c/d/e/f/g.xml", "../../../file.xml"),
                "http://example.com/a/b/c/file.xml",
            ),
        ]
        for i, elem in enumerate(test_arr):
            # only windows uses the \\ file path separator
            # for now skip the first tests with \\ if we are on a unix system, since the \\ is an invalid path on
            # a unix like os such as macOS or linux
            if elem[0][0].startswith("E:\\") and os.sep != "\\":
                logging.info("Skipping Windows specific unit test case")
                continue
            expected = elem[1]
            received = resolve_uri(elem[0][0], elem[0][1])
            self.assertEqual(expected, received, msg=f"Failed at test elem {i}")

    def test_compare_uri(self):
        test_arr = [
            ["./abc", "abc", True],
            ["./abc", "\\abc\\", True],
            ["./abc", "abcd", False],
            ["http://abc.de", "https://abc.de", True],
        ]
        for i, test_case in enumerate(test_arr):
            expected = test_case[2]
            received = compare_uri(test_case[0], test_case[1])
            self.assertEqual(expected, received, msg=f"Failed at test elem {i}")


if __name__ == "__main__":
    unittest.main()



"""
This file holds a fallback namespace map for common XBRL taxonomies

Normally, the taxonomy schema should define the namespaces used and provide their schemaUrl, however some do not.
For this reason we have this fallback map to help resolve common taxonomies.
"""

NS_MAP: dict[str, str] = {
    "http://arelle.org/doc/2014-01-31": "http://arelle.org/2014/doc-2014-01-31.xsd",
    "http://fasb.org/dis/cecltmp01/2019-01-31": "http://xbrl.fasb.org/us-gaap/2019/dis/us-gaap-dis-cecltmp01-2019-01-31.xsd",
    "http://fasb.org/dis/cecltmp02/2019-01-31": "http://xbrl.fasb.org/us-gaap/2019/dis/us-gaap-dis-cecltmp02-2019-01-31.xsd",
    "http://fasb.org/dis/cecltmp03/2019-01-31": "http://xbrl.fasb.org/us-gaap/2019/dis/us-gaap-dis-cecltmp03-2019-01-31.xsd",
    "http://fasb.org/dis/cecltmp04/2019-01-31": "http://xbrl.fasb.org/us-gaap/2019/dis/us-gaap-dis-cecltmp04-2019-01-31.xsd",
    "http://fasb.org/dis/cecltmp05/2019-01-31": "http://xbrl.fasb.org/us-gaap/2019/dis/us-gaap-dis-cecltmp05-2019-01-31.xsd",
    "http://fasb.org/dis/fifvdtmp01/2018-01-31": "http://xbrl.fasb.org/us-gaap/2018/dis/us-gaap-dis-fifvdtmp01-2018-01-31.xsd",
    "http://fasb.org/dis/fifvdtmp01/2019-01-31": "http://xbrl.fasb.org/us-gaap/2019/dis/us-gaap-dis-fifvdtmp01-2019-01-31.xsd",
    "http://fasb.org/dis/fifvdtmp02/2018-01-31": "http://xbrl.fasb.org/us-gaap/2018/dis/us-gaap-dis-fifvdtmp02-2018-01-31.xsd",
    "http://fasb.org/dis/fifvdtmp02/2019-01-31": "http://xbrl.fasb.org/us-gaap/2019/dis/us-gaap-dis-fifvdtmp02-2019-01-31.xsd",
    "http://fasb.org/dis/idestmp011/2018-01-31": "http://xbrl.fasb.org/us-gaap/2018/dis/us-gaap-dis-idestmp011-2018-01-31.xsd",
    "http://fasb.org/dis/idestmp011/2019-01-31": "http://xbrl.fasb.org/us-gaap/2019/dis/us-gaap-dis-idestmp011-2019-01-31.xsd",
    "http://fasb.org/dis/idestmp012/2018-01-31": "http://xbrl.fasb.org/us-gaap/2018/dis/us-gaap-dis-idestmp012-2018-01-31.xsd",
    "http://fasb.org/dis/idestmp012/2019-01-31": "http://xbrl.fasb.org/us-gaap/2019/dis/us-gaap-dis-idestmp012-2019-01-31.xsd",
    "http://fasb.org/dis/idestmp021/2018-01-31": "http://xbrl.fasb.org/us-gaap/2018/dis/us-gaap-dis-idestmp021-2018-01-31.xsd",
    "http://fasb.org/dis/idestmp021/2019-01-31": "http://xbrl.fasb.org/us-gaap/2019/dis/us-gaap-dis-idestmp021-2019-01-31.xsd",
    "http://fasb.org/dis/idestmp022/2018-01-31": "http://xbrl.fasb.org/us-gaap/2018/dis/us-gaap-dis-idestmp022-2018-01-31.xsd",
    "http://fasb.org/dis/idestmp022/2019-01-31": "http://xbrl.fasb.org/us-gaap/2019/dis/us-gaap-dis-idestmp022-2019-01-31.xsd",
    "http://fasb.org/dis/idestmp03/2018-01-31": "http://xbrl.fasb.org/us-gaap/2018/dis/us-gaap-dis-idestmp03-2018-01-31.xsd",
    "http://fasb.org/dis/idestmp03/2019-01-31": "http://xbrl.fasb.org/us-gaap/2019/dis/us-gaap-dis-idestmp03-2019-01-31.xsd",
    "http://fasb.org/dis/idestmp04/2018-01-31": "http://xbrl.fasb.org/us-gaap/2018/dis/us-gaap-dis-idestmp04-2018-01-31.xsd",
    "http://fasb.org/dis/idestmp04/2019-01-31": "http://xbrl.fasb.org/us-gaap/2019/dis/us-gaap-dis-idestmp04-2019-01-31.xsd",
    "http://fasb.org/dis/insldtmp01/2019-01-31": "http://xbrl.fasb.org/us-gaap/2019/dis/us-gaap-dis-insldtmp01-2019-01-31.xsd",
    "http://fasb.org/dis/insldtmp021/2019-01-31": "http://xbrl.fasb.org/us-gaap/2019/dis/us-gaap-dis-insldtmp021-2019-01-31.xsd",
    "http://fasb.org/dis/insldtmp022/2019-01-31": "http://xbrl.fasb.org/us-gaap/2019/dis/us-gaap-dis-insldtmp022-2019-01-31.xsd",
    "http://fasb.org/dis/insldtmp023/2019-01-31": "http://xbrl.fasb.org/us-gaap/2019/dis/us-gaap-dis-insldtmp023-2019-01-31.xsd",
    "http://fasb.org/dis/insldtmp024/2019-01-31": "http://xbrl.fasb.org/us-gaap/2019/dis/us-gaap-dis-insldtmp024-2019-01-31.xsd",
    "http://fasb.org/dis/insldtmp025/2019-01-31": "http://xbrl.fasb.org/us-gaap/2019/dis/us-gaap-dis-insldtmp025-2019-01-31.xsd",
    "http://fasb.org/dis/insldtmp031/2019-01-31": "http://xbrl.fasb.org/us-gaap/2019/dis/us-gaap-dis-insldtmp031-2019-01-31.xsd",
    "http://fasb.org/dis/insldtmp032/2019-01-31": "http://xbrl.fasb.org/us-gaap/2019/dis/us-gaap-dis-insldtmp032-2019-01-31.xsd",
    "http://fasb.org/dis/insldtmp033/2019-01-31": "http://xbrl.fasb.org/us-gaap/2019/dis/us-gaap-dis-insldtmp033-2019-01-31.xsd",
    "http://fasb.org/dis/insldtmp041/2019-01-31": "http://xbrl.fasb.org/us-gaap/2019/dis/us-gaap-dis-insldtmp041-2019-01-31.xsd",
    "http://fasb.org/dis/insldtmp042/2019-01-31": "http://xbrl.fasb.org/us-gaap/2019/dis/us-gaap-dis-insldtmp042-2019-01-31.xsd",
    "http://fasb.org/dis/insldtmp051/2019-01-31": "http://xbrl.fasb.org/us-gaap/2019/dis/us-gaap-dis-insldtmp051-2019-01-31.xsd",
    "http://fasb.org/dis/insldtmp052/2019-01-31": "http://xbrl.fasb.org/us-gaap/2019/dis/us-gaap-dis-insldtmp052-2019-01-31.xsd",
    "http://fasb.org/dis/insldtmp061/2019-01-31": "http://xbrl.fasb.org/us-gaap/2019/dis/us-gaap-dis-insldtmp061-2019-01-31.xsd",
    "http://fasb.org/dis/insldtmp062/2019-01-31": "http://xbrl.fasb.org/us-gaap/2019/dis/us-gaap-dis-insldtmp062-2019-01-31.xsd",
    "http://fasb.org/dis/leasestmp01/2017-01-31": "http://xbrl.fasb.org/us-gaap/2017/dis/us-gaap-dis-leasestmp01-2017-01-31.xsd",
    "http://fasb.org/dis/leasestmp02/2017-01-31": "http://xbrl.fasb.org/us-gaap/2017/dis/us-gaap-dis-leasestmp02-2017-01-31.xsd",
    "http://fasb.org/dis/leasestmp03/2017-01-31": "http://xbrl.fasb.org/us-gaap/2017/dis/us-gaap-dis-leasestmp03-2017-01-31.xsd",
    "http://fasb.org/dis/leasestmp04/2017-01-31": "http://xbrl.fasb.org/us-gaap/2017/dis/us-gaap-dis-leasestmp04-2017-01-31.xsd",
    "http://fasb.org/dis/leasestmp05/2017-01-31": "http://xbrl.fasb.org/us-gaap/2017/dis/us-gaap-dis-leasestmp05-2017-01-31.xsd",
    "http://fasb.org/dis/leastmp01/2018-01-31": "http://xbrl.fasb.org/us-gaap/2018/dis/us-gaap-dis-leastmp01-2018-01-31.xsd",
    "http://fasb.org/dis/leastmp01/2019-01-31": "http://xbrl.fasb.org/us-gaap/2019/dis/us-gaap-dis-leastmp01-2019-01-31.xsd",
    "http://fasb.org/dis/leastmp02/2018-01-31": "http://xbrl.fasb.org/us-gaap/2018/dis/us-gaap-dis-leastmp02-2018-01-31.xsd",
    "http://fasb.org/dis/leastmp02/2019-01-31": "http://xbrl.fasb.org/us-gaap/2019/dis/us-gaap-dis-leastmp02-2019-01-31.xsd",
    "http://fasb.org/dis/leastmp03/2018-01-31": "http://xbrl.fasb.org/us-gaap/2018/dis/us-gaap-dis-leastmp03-2018-01-31.xsd",
    "http://fasb.org/dis/leastmp03/2019-01-31": "http://xbrl.fasb.org/us-gaap/2019/dis/us-gaap-dis-leastmp03-2019-01-31.xsd",
    "http://fasb.org/dis/leastmp04/2018-01-31": "http://xbrl.fasb.org/us-gaap/2018/dis/us-gaap-dis-leastmp04-2018-01-31.xsd",
    "http://fasb.org/dis/leastmp04/2019-01-31": "http://xbrl.fasb.org/us-gaap/2019/dis/us-gaap-dis-leastmp04-2019-01-31.xsd",
    "http://fasb.org/dis/leastmp05/2018-01-31": "http://xbrl.fasb.org/us-gaap/2018/dis/us-gaap-dis-leastmp05-2018-01-31.xsd",
    "http://fasb.org/dis/leastmp05/2019-01-31": "http://xbrl.fasb.org/us-gaap/2019/dis/us-gaap-dis-leastmp05-2019-01-31.xsd",
    "http://fasb.org/dis/rbtmp011/2017-01-31": "http://xbrl.fasb.org/us-gaap/2017/dis/us-gaap-dis-rbtmp011-2017-01-31.xsd",
    "http://fasb.org/dis/rbtmp011/2018-01-31": "http://xbrl.fasb.org/us-gaap/2018/dis/us-gaap-dis-rbtmp011-2018-01-31.xsd",
    "http://fasb.org/dis/rbtmp011/2019-01-31": "http://xbrl.fasb.org/us-gaap/2019/dis/us-gaap-dis-rbtmp011-2019-01-31.xsd",
    "http://fasb.org/dis/rbtmp012/2017-01-31": "http://xbrl.fasb.org/us-gaap/2017/dis/us-gaap-dis-rbtmp012-2017-01-31.xsd",
    "http://fasb.org/dis/rbtmp012/2018-01-31": "http://xbrl.fasb.org/us-gaap/2018/dis/us-gaap-dis-rbtmp012-2018-01-31.xsd",
    "http://fasb.org/dis/rbtmp012/2019-01-31": "http://xbrl.fasb.org/us-gaap/2019/dis/us-gaap-dis-rbtmp012-2019-01-31.xsd",
    "http://fasb.org/dis/rbtmp02/2017-01-31": "http://xbrl.fasb.org/us-gaap/2017/dis/us-gaap-dis-rbtmp02-2017-01-31.xsd",
    "http://fasb.org/dis/rbtmp02/2018-01-31": "http://xbrl.fasb.org/us-gaap/2018/dis/us-gaap-dis-rbtmp02-2018-01-31.xsd",
    "http://fasb.org/dis/rbtmp02/2019-01-31": "http://xbrl.fasb.org/us-gaap/2019/dis/us-gaap-dis-rbtmp02-2019-01-31.xsd",
    "http://fasb.org/dis/rbtmp03/2017-01-31": "http://xbrl.fasb.org/us-gaap/2017/dis/us-gaap-dis-rbtmp03-2017-01-31.xsd",
    "http://fasb.org/dis/rbtmp03/2018-01-31": "http://xbrl.fasb.org/us-gaap/2018/dis/us-gaap-dis-rbtmp03-2018-01-31.xsd",
    "http://fasb.org/dis/rbtmp03/2019-01-31": "http://xbrl.fasb.org/us-gaap/2019/dis/us-gaap-dis-rbtmp03-2019-01-31.xsd",
    "http://fasb.org/dis/rbtmp04/2017-01-31": "http://xbrl.fasb.org/us-gaap/2017/dis/us-gaap-dis-rbtmp04-2017-01-31.xsd",
    "http://fasb.org/dis/rbtmp04/2018-01-31": "http://xbrl.fasb.org/us-gaap/2018/dis/us-gaap-dis-rbtmp04-2018-01-31.xsd",
    "http://fasb.org/dis/rbtmp04/2019-01-31": "http://xbrl.fasb.org/us-gaap/2019/dis/us-gaap-dis-rbtmp04-2019-01-31.xsd",
    "http://fasb.org/dis/rbtmp041/2017-01-31": "http://xbrl.fasb.org/us-gaap/2017/dis/us-gaap-dis-rbtmp041-2017-01-31.xsd",
    "http://fasb.org/dis/rbtmp041/2018-01-31": "http://xbrl.fasb.org/us-gaap/2018/dis/us-gaap-dis-rbtmp041-2018-01-31.xsd",
    "http://fasb.org/dis/rbtmp041/2019-01-31": "http://xbrl.fasb.org/us-gaap/2019/dis/us-gaap-dis-rbtmp041-2019-01-31.xsd",
    "http://fasb.org/dis/rbtmp05/2017-01-31": "http://xbrl.fasb.org/us-gaap/2017/dis/us-gaap-dis-rbtmp05-2017-01-31.xsd",
    "http://fasb.org/dis/rbtmp05/2018-01-31": "http://xbrl.fasb.org/us-gaap/2018/dis/us-gaap-dis-rbtmp05-2018-01-31.xsd",
    "http://fasb.org/dis/rbtmp05/2019-01-31": "http://xbrl.fasb.org/us-gaap/2019/dis/us-gaap-dis-rbtmp05-2019-01-31.xsd",
    "http://fasb.org/dis/rbtmp06/2017-01-31": "http://xbrl.fasb.org/us-gaap/2017/dis/us-gaap-dis-rbtmp06-2017-01-31.xsd",
    "http://fasb.org/dis/rbtmp06/2018-01-31": "http://xbrl.fasb.org/us-gaap/2018/dis/us-gaap-dis-rbtmp06-2018-01-31.xsd",
    "http://fasb.org/dis/rbtmp06/2019-01-31": "http://xbrl.fasb.org/us-gaap/2019/dis/us-gaap-dis-rbtmp06-2019-01-31.xsd",
    "http://fasb.org/dis/rbtmp07/2017-01-31": "http://xbrl.fasb.org/us-gaap/2017/dis/us-gaap-dis-rbtmp07-2017-01-31.xsd",
    "http://fasb.org/dis/rbtmp07/2018-01-31": "http://xbrl.fasb.org/us-gaap/2018/dis/us-gaap-dis-rbtmp07-2018-01-31.xsd",
    "http://fasb.org/dis/rbtmp07/2019-01-31": "http://xbrl.fasb.org/us-gaap/2019/dis/us-gaap-dis-rbtmp07-2019-01-31.xsd",
    "http://fasb.org/dis/rbtmp08/2017-01-31": "http://xbrl.fasb.org/us-gaap/2017/dis/us-gaap-dis-rbtmp08-2017-01-31.xsd",
    "http://fasb.org/dis/rbtmp08/2018-01-31": "http://xbrl.fasb.org/us-gaap/2018/dis/us-gaap-dis-rbtmp08-2018-01-31.xsd",
    "http://fasb.org/dis/rbtmp08/2019-01-31": "http://xbrl.fasb.org/us-gaap/2019/dis/us-gaap-dis-rbtmp08-2019-01-31.xsd",
    "http://fasb.org/dis/rbtmp09/2017-01-31": "http://xbrl.fasb.org/us-gaap/2017/dis/us-gaap-dis-rbtmp09-2017-01-31.xsd",
    "http://fasb.org/dis/rbtmp09/2018-01-31": "http://xbrl.fasb.org/us-gaap/2018/dis/us-gaap-dis-rbtmp09-2018-01-31.xsd",
    "http://fasb.org/dis/rbtmp09/2019-01-31": "http://xbrl.fasb.org/us-gaap/2019/dis/us-gaap-dis-rbtmp09-2019-01-31.xsd",
    "http://fasb.org/dis/rbtmp102/2017-01-31": "http://xbrl.fasb.org/us-gaap/2017/dis/us-gaap-dis-rbtmp102-2017-01-31.xsd",
    "http://fasb.org/dis/rbtmp102/2018-01-31": "http://xbrl.fasb.org/us-gaap/2018/dis/us-gaap-dis-rbtmp102-2018-01-31.xsd",
    "http://fasb.org/dis/rbtmp102/2019-01-31": "http://xbrl.fasb.org/us-gaap/2019/dis/us-gaap-dis-rbtmp102-2019-01-31.xsd",
    "http://fasb.org/dis/rbtmp103/2017-01-31": "http://xbrl.fasb.org/us-gaap/2017/dis/us-gaap-dis-rbtmp103-2017-01-31.xsd",
    "http://fasb.org/dis/rbtmp103/2018-01-31": "http://xbrl.fasb.org/us-gaap/2018/dis/us-gaap-dis-rbtmp103-2018-01-31.xsd",
    "http://fasb.org/dis/rbtmp103/2019-01-31": "http://xbrl.fasb.org/us-gaap/2019/dis/us-gaap-dis-rbtmp103-2019-01-31.xsd",
    "http://fasb.org/dis/rbtmp104/2017-01-31": "http://xbrl.fasb.org/us-gaap/2017/dis/us-gaap-dis-rbtmp104-2017-01-31.xsd",
    "http://fasb.org/dis/rbtmp104/2018-01-31": "http://xbrl.fasb.org/us-gaap/2018/dis/us-gaap-dis-rbtmp104-2018-01-31.xsd",
    "http://fasb.org/dis/rbtmp104/2019-01-31": "http://xbrl.fasb.org/us-gaap/2019/dis/us-gaap-dis-rbtmp104-2019-01-31.xsd",
    "http://fasb.org/dis/rbtmp105/2017-01-31": "http://xbrl.fasb.org/us-gaap/2017/dis/us-gaap-dis-rbtmp105-2017-01-31.xsd",
    "http://fasb.org/dis/rbtmp105/2018-01-31": "http://xbrl.fasb.org/us-gaap/2018/dis/us-gaap-dis-rbtmp105-2018-01-31.xsd",
    "http://fasb.org/dis/rbtmp105/2019-01-31": "http://xbrl.fasb.org/us-gaap/2019/dis/us-gaap-dis-rbtmp105-2019-01-31.xsd",
    "http://fasb.org/dis/rbtmp111/2017-01-31": "http://xbrl.fasb.org/us-gaap/2017/dis/us-gaap-dis-rbtmp111-2017-01-31.xsd",
    "http://fasb.org/dis/rbtmp111/2018-01-31": "http://xbrl.fasb.org/us-gaap/2018/dis/us-gaap-dis-rbtmp111-2018-01-31.xsd",
    "http://fasb.org/dis/rbtmp111/2019-01-31": "http://xbrl.fasb.org/us-gaap/2019/dis/us-gaap-dis-rbtmp111-2019-01-31.xsd",
    "http://fasb.org/dis/rbtmp112/2017-01-31": "http://xbrl.fasb.org/us-gaap/2017/dis/us-gaap-dis-rbtmp112-2017-01-31.xsd",
    "http://fasb.org/dis/rbtmp112/2018-01-31": "http://xbrl.fasb.org/us-gaap/2018/dis/us-gaap-dis-rbtmp112-2018-01-31.xsd",
    "http://fasb.org/dis/rbtmp112/2019-01-31": "http://xbrl.fasb.org/us-gaap/2019/dis/us-gaap-dis-rbtmp112-2019-01-31.xsd",
    "http://fasb.org/dis/rbtmp121/2017-01-31": "http://xbrl.fasb.org/us-gaap/2017/dis/us-gaap-dis-rbtmp121-2017-01-31.xsd",
    "http://fasb.org/dis/rbtmp121/2018-01-31": "http://xbrl.fasb.org/us-gaap/2018/dis/us-gaap-dis-rbtmp121-2018-01-31.xsd",
    "http://fasb.org/dis/rbtmp121/2019-01-31": "http://xbrl.fasb.org/us-gaap/2019/dis/us-gaap-dis-rbtmp121-2019-01-31.xsd",
    "http://fasb.org/dis/rbtmp122/2017-01-31": "http://xbrl.fasb.org/us-gaap/2017/dis/us-gaap-dis-rbtmp122-2017-01-31.xsd",
    "http://fasb.org/dis/rbtmp122/2018-01-31": "http://xbrl.fasb.org/us-gaap/2018/dis/us-gaap-dis-rbtmp122-2018-01-31.xsd",
    "http://fasb.org/dis/rbtmp122/2019-01-31": "http://xbrl.fasb.org/us-gaap/2019/dis/us-gaap-dis-rbtmp122-2019-01-31.xsd",
    "http://fasb.org/dis/rbtmp123/2017-01-31": "http://xbrl.fasb.org/us-gaap/2017/dis/us-gaap-dis-rbtmp123-2017-01-31.xsd",
    "http://fasb.org/dis/rbtmp123/2018-01-31": "http://xbrl.fasb.org/us-gaap/2018/dis/us-gaap-dis-rbtmp123-2018-01-31.xsd",
    "http://fasb.org/dis/rbtmp123/2019-01-31": "http://xbrl.fasb.org/us-gaap/2019/dis/us-gaap-dis-rbtmp123-2019-01-31.xsd",
    "http://fasb.org/dis/rbtmp125/2017-01-31": "http://xbrl.fasb.org/us-gaap/2017/dis/us-gaap-dis-rbtmp125-2017-01-31.xsd",
    "http://fasb.org/dis/rbtmp125/2018-01-31": "http://xbrl.fasb.org/us-gaap/2018/dis/us-gaap-dis-rbtmp125-2018-01-31.xsd",
    "http://fasb.org/dis/rbtmp125/2019-01-31": "http://xbrl.fasb.org/us-gaap/2019/dis/us-gaap-dis-rbtmp125-2019-01-31.xsd",
    "http://fasb.org/dis/rbtmp131/2017-01-31": "http://xbrl.fasb.org/us-gaap/2017/dis/us-gaap-dis-rbtmp131-2017-01-31.xsd",
    "http://fasb.org/dis/rbtmp131/2018-01-31": "http://xbrl.fasb.org/us-gaap/2018/dis/us-gaap-dis-rbtmp131-2018-01-31.xsd",
    "http://fasb.org/dis/rbtmp131/2019-01-31": "http://xbrl.fasb.org/us-gaap/2019/dis/us-gaap-dis-rbtmp131-2019-01-31.xsd",
    "http://fasb.org/dis/rbtmp141/2017-01-31": "http://xbrl.fasb.org/us-gaap/2017/dis/us-gaap-dis-rbtmp141-2017-01-31.xsd",
    "http://fasb.org/dis/rbtmp141/2018-01-31": "http://xbrl.fasb.org/us-gaap/2018/dis/us-gaap-dis-rbtmp141-2018-01-31.xsd",
    "http://fasb.org/dis/rbtmp141/2019-01-31": "http://xbrl.fasb.org/us-gaap/2019/dis/us-gaap-dis-rbtmp141-2019-01-31.xsd",
    "http://fasb.org/dis/rcctmp01/2017-01-31": "http://xbrl.fasb.org/us-gaap/2017/dis/us-gaap-dis-rcctmp01-2017-01-31.xsd",
    "http://fasb.org/dis/rcctmp01/2018-01-31": "http://xbrl.fasb.org/us-gaap/2018/dis/us-gaap-dis-rcctmp01-2018-01-31.xsd",
    "http://fasb.org/dis/rcctmp01/2019-01-31": "http://xbrl.fasb.org/us-gaap/2019/dis/us-gaap-dis-rcctmp01-2019-01-31.xsd",
    "http://fasb.org/dis/rcctmp03/2017-01-31": "http://xbrl.fasb.org/us-gaap/2017/dis/us-gaap-dis-rcctmp03-2017-01-31.xsd",
    "http://fasb.org/dis/rcctmp03/2018-01-31": "http://xbrl.fasb.org/us-gaap/2018/dis/us-gaap-dis-rcctmp03-2018-01-31.xsd",
    "http://fasb.org/dis/rcctmp03/2019-01-31": "http://xbrl.fasb.org/us-gaap/2019/dis/us-gaap-dis-rcctmp03-2019-01-31.xsd",
    "http://fasb.org/dis/rcctmp04/2017-01-31": "http://xbrl.fasb.org/us-gaap/2017/dis/us-gaap-dis-rcctmp04-2017-01-31.xsd",
    "http://fasb.org/dis/rcctmp04/2018-01-31": "http://xbrl.fasb.org/us-gaap/2018/dis/us-gaap-dis-rcctmp04-2018-01-31.xsd",
    "http://fasb.org/dis/rcctmp04/2019-01-31": "http://xbrl.fasb.org/us-gaap/2019/dis/us-gaap-dis-rcctmp04-2019-01-31.xsd",
    "http://fasb.org/dis/rcctmp05/2017-01-31": "http://xbrl.fasb.org/us-gaap/2017/dis/us-gaap-dis-rcctmp05-2017-01-31.xsd",
    "http://fasb.org/dis/rcctmp05/2018-01-31": "http://xbrl.fasb.org/us-gaap/2018/dis/us-gaap-dis-rcctmp05-2018-01-31.xsd",
    "http://fasb.org/dis/rcctmp05/2019-01-31": "http://xbrl.fasb.org/us-gaap/2019/dis/us-gaap-dis-rcctmp05-2019-01-31.xsd",
    "http://fasb.org/srt/2018-01-31": "http://xbrl.fasb.org/srt/2018/elts/srt-2018-01-31.xsd",
    "http://fasb.org/srt/2019-01-31": "http://xbrl.fasb.org/srt/2019/elts/srt-2019-01-31.xsd",
    "http://fasb.org/srt/2020-01-31": "http://xbrl.fasb.org/srt/2020/elts/srt-2020-01-31.xsd",
    "http://fasb.org/srt/2021-01-31": "http://xbrl.fasb.org/srt/2021/elts/srt-2021-01-31.xsd",
    "http://fasb.org/srt/2022": "https://xbrl.fasb.org/srt/2022/elts/srt-2022.xsd",
    "http://fasb.org/srt/2023": "https://xbrl.fasb.org/srt/2023/elts/srt-2023.xsd",
    "http://fasb.org/srt/2024": "https://xbrl.fasb.org/srt/2024/elts/srt-2024.xsd",
    "http://fasb.org/srt/2025": "https://xbrl.fasb.org/srt/2025/elts/srt-2025.xsd",
    "http://fasb.org/srt-roles/2018-01-31": "http://xbrl.fasb.org/srt/2018/elts/srt-roles-2018-01-31.xsd",
    "http://fasb.org/srt-roles/2019-01-31": "http://xbrl.fasb.org/srt/2019/elts/srt-roles-2019-01-31.xsd",
    "http://fasb.org/srt-roles/2020-01-31": "http://xbrl.fasb.org/srt/2020/elts/srt-roles-2020-01-31.xsd",
    "http://fasb.org/srt-roles/2021-01-31": "http://xbrl.fasb.org/srt/2021/elts/srt-roles-2021-01-31.xsd",
    "http://fasb.org/srt-types/2018-01-31": "http://xbrl.fasb.org/srt/2018/elts/srt-types-2018-01-31.xsd",
    "http://fasb.org/srt-types/2019-01-31": "http://xbrl.fasb.org/srt/2019/elts/srt-types-2019-01-31.xsd",
    "http://fasb.org/srt-types/2020-01-31": "http://xbrl.fasb.org/srt/2020/elts/srt-types-2020-01-31.xsd",
    "http://fasb.org/srt-types/2021-01-31": "http://xbrl.fasb.org/srt/2021/elts/srt-types-2021-01-31.xsd",
    "http://fasb.org/us-gaap/2011-01-31": "http://xbrl.fasb.org/us-gaap/2011/elts/us-gaap-2011-01-31.xsd",
    "http://fasb.org/us-gaap/2012-01-31": "http://xbrl.fasb.org/us-gaap/2012/elts/us-gaap-2012-01-31.xsd",
    "http://fasb.org/us-gaap/2013-01-31": "http://xbrl.fasb.org/us-gaap/2013/elts/us-gaap-2013-01-31.xsd",
    "http://fasb.org/us-gaap/2014-01-31": "http://xbrl.fasb.org/us-gaap/2014/elts/us-gaap-2014-01-31.xsd",
    "http://fasb.org/us-gaap/2015-01-31": "http://xbrl.fasb.org/us-gaap/2015/elts/us-gaap-2015-01-31.xsd",
    "http://fasb.org/us-gaap/2016-01-31": "http://xbrl.fasb.org/us-gaap/2016/elts/us-gaap-2016-01-31.xsd",
    "http://fasb.org/us-gaap/2017-01-31": "http://xbrl.fasb.org/us-gaap/2017/elts/us-gaap-2017-01-31.xsd",
    "http://fasb.org/us-gaap/2018-01-31": "http://xbrl.fasb.org/us-gaap/2018/elts/us-gaap-2018-01-31.xsd",
    "http://fasb.org/us-gaap/2019-01-31": "http://xbrl.fasb.org/us-gaap/2019/elts/us-gaap-2019-01-31.xsd",
    "http://fasb.org/us-gaap/2020-01-31": "http://xbrl.fasb.org/us-gaap/2020/elts/us-gaap-2020-01-31.xsd",
    "http://fasb.org/us-gaap/2021-01-31": "http://xbrl.fasb.org/us-gaap/2021/elts/us-gaap-2021-01-31.xsd",
    "http://fasb.org/us-gaap/2022-01-31": "https://xbrl.fasb.org/us-gaap/2022/elts/us-gaap-2022.xsd",
    "http://fasb.org/us-gaap/2022": "https://xbrl.fasb.org/us-gaap/2022/elts/us-gaap-2022.xsd",
    "http://fasb.org/us-gaap/2023": "https://xbrl.fasb.org/us-gaap/2023/elts/us-gaap-2023.xsd",
    "http://fasb.org/us-gaap/2024": "https://xbrl.fasb.org/us-gaap/2024/elts/us-gaap-2024.xsd",
    "http://fasb.org/us-gaap/2025": "https://xbrl.fasb.org/us-gaap/2025/elts/us-gaap-2025.xsd",
    "http://fasb.org/us-roles/2011-01-31": "http://xbrl.fasb.org/us-gaap/2011/elts/us-roles-2011-01-31.xsd",
    "http://fasb.org/us-roles/2012-01-31": "http://xbrl.fasb.org/us-gaap/2012/elts/us-roles-2012-01-31.xsd",
    "http://fasb.org/us-roles/2013-01-31": "http://xbrl.fasb.org/us-gaap/2013/elts/us-roles-2013-01-31.xsd",
    "http://fasb.org/us-roles/2014-01-31": "http://xbrl.fasb.org/us-gaap/2014/elts/us-roles-2014-01-31.xsd",
    "http://fasb.org/us-roles/2015-01-31": "http://xbrl.fasb.org/us-gaap/2015/elts/us-roles-2015-01-31.xsd",
    "http://fasb.org/us-roles/2016-01-31": "http://xbrl.fasb.org/us-gaap/2016/elts/us-roles-2016-01-31.xsd",
    "http://fasb.org/us-roles/2017-01-31": "http://xbrl.fasb.org/us-gaap/2017/elts/us-roles-2017-01-31.xsd",
    "http://fasb.org/us-roles/2018-01-31": "http://xbrl.fasb.org/us-gaap/2018/elts/us-roles-2018-01-31.xsd",
    "http://fasb.org/us-roles/2019-01-31": "http://xbrl.fasb.org/us-gaap/2019/elts/us-roles-2019-01-31.xsd",
    "http://fasb.org/us-roles/2020-01-31": "http://xbrl.fasb.org/us-gaap/2020/elts/us-roles-2020-01-31.xsd",
    "http://fasb.org/us-roles/2021-01-31": "http://xbrl.fasb.org/us-gaap/2021/elts/us-roles-2021-01-31.xsd",
    "http://fasb.org/us-types/2011-01-31": "http://xbrl.fasb.org/us-gaap/2011/elts/us-types-2011-01-31.xsd",
    "http://fasb.org/us-types/2012-01-31": "http://xbrl.fasb.org/us-gaap/2012/elts/us-types-2012-01-31.xsd",
    "http://fasb.org/us-types/2013-01-31": "http://xbrl.fasb.org/us-gaap/2013/elts/us-types-2013-01-31.xsd",
    "http://fasb.org/us-types/2014-01-31": "http://xbrl.fasb.org/us-gaap/2014/elts/us-types-2014-01-31.xsd",
    "http://fasb.org/us-types/2015-01-31": "http://xbrl.fasb.org/us-gaap/2015/elts/us-types-2015-01-31.xsd",
    "http://fasb.org/us-types/2016-01-31": "http://xbrl.fasb.org/us-gaap/2016/elts/us-types-2016-01-31.xsd",
    "http://fasb.org/us-types/2017-01-31": "http://xbrl.fasb.org/us-gaap/2017/elts/us-types-2017-01-31.xsd",
    "http://fasb.org/us-types/2018-01-31": "http://xbrl.fasb.org/us-gaap/2018/elts/us-types-2018-01-31.xsd",
    "http://fasb.org/us-types/2019-01-31": "http://xbrl.fasb.org/us-gaap/2020/elts/us-types-2020-01-31.xsd",
    "http://fasb.org/us-types/2021-01-31": "http://xbrl.fasb.org/us-gaap/2021/elts/us-types-2021-01-31.xsd",
    "http://ici.org/rr/2006": "http://xbrl.ici.org/rr/2006/ici-rr.xsd",
    "http://www.esma.europa.eu/xbrl/esef/arcrole/wider-narrower": "http://www.xbrl.org/lrr/arcrole/esma-arcrole-2018-11-21.xsd",
    "http://www.w3.org/1999/xlink": "http://www.xbrl.org/2003/xlink-2003-12-31.xsd",
    "http://www.xbrl.org/2003/instance": "http://www.xbrl.org/2003/xbrl-instance-2003-12-31.xsd",
    "http://www.xbrl.org/2003/linkbase": "http://www.xbrl.org/2003/xbrl-linkbase-2003-12-31.xsd",
    "http://www.xbrl.org/2003/XLink": "http://www.xbrl.org/2003/xl-2003-12-31.xsd",
    "http://www.xbrl.org/2004/ref": "http://www.xbrl.org/2004/ref-2004-08-10.xsd",
    "http://www.xbrl.org/2006/ref": "http://www.xbrl.org/2006/ref-2006-02-27.xsd",
    "http://www.xbrl.org/2009/arcrole/deprecated": "http://www.xbrl.org/lrr/arcrole/deprecated-2009-12-16.xsd",
    "http://www.xbrl.org/2009/arcrole/fact-explanatoryFact": "http://www.xbrl.org/lrr/arcrole/factExplanatory-2009-12-16.xsd",
    "http://www.xbrl.org/2009/role/deprecated": "http://www.xbrl.org/lrr/role/deprecated-2009-12-16.xsd",
    "http://www.xbrl.org/2009/role/negated": "http://www.xbrl.org/lrr/role/negated-2009-12-16.xsd",
    "http://www.xbrl.org/2009/role/net": "http://www.xbrl.org/lrr/role/net-2009-12-16.xsd",
    "http://www.xbrl.org/dtr/type/2020-01-21": "http://www.xbrl.org/dtr/type/2020-01-21/types.xsd",
    "http://www.xbrl.org/dtr/type/non-numeric": "http://www.xbrl.org/dtr/type/nonNumeric-2009-12-20.xsd",
    "http://www.xbrl.org/dtr/type/numeric": "http://www.xbrl.org/dtr/type/numeric-2009-12-16.xsd",
    "http://www.xbrl.org/us/fr/common/fste/2005-02-28": "http://www.xbrl.org/us/fr/common/fste/2005-02-28/usfr-fste-2005-02-28.xsd",
    "http://www.xbrl.org/us/fr/common/fstr/2005-02-28": "http://www.xbrl.org/us/fr/common/fstr/2005-02-28/usfr-fstr-2005-02-28.xsd",
    "http://www.xbrl.org/us/fr/common/ime/2005-06-28": "http://www.xbrl.org/us/fr/common/ime/2005-06-28/usfr-ime-2005-06-28.xsd",
    "http://www.xbrl.org/us/fr/common/pte/2005-02-28": "http://www.xbrl.org/us/fr/common/pte/2005-02-28/usfr-pte-2005-02-28.xsd",
    "http://www.xbrl.org/us/fr/common/ptr/2005-02-28": "http://www.xbrl.org/us/fr/common/ptr/2005-02-28/usfr-ptr-2005-02-28.xsd",
    "http://www.xbrl.org/us/fr/gaap/basi/2005-02-28": "http://www.xbrl.org/us/fr/gaap/basi/2005-02-28/us-gaap-basi-2005-02-28.xsd",
    "http://www.xbrl.org/us/fr/gaap/ci/2005-02-28": "http://www.xbrl.org/us/fr/gaap/ci/2005-02-28/us-gaap-ci-2005-02-28.xsd",
    "http://www.xbrl.org/us/fr/gaap/im/2005-06-28": "http://www.xbrl.org/us/fr/gaap/im/2005-06-28/us-gaap-im-2005-06-28.xsd",
    "http://www.xbrl.org/us/fr/gaap/ins/2005-02-28": "http://www.xbrl.org/us/fr/gaap/ins/2005-02-28/us-gaap-ins-2005-02-28.xsd",
    "http://www.xbrl.org/us/fr/rpt/ar/2005-02-28": "http://www.xbrl.org/us/fr/rpt/ar/2005-02-28/usfr-ar-2005-02-28.xsd",
    "http://www.xbrl.org/us/fr/rpt/mda/2005-02-28": "http://www.xbrl.org/us/fr/rpt/mda/2005-02-28/usfr-mda-2005-02-28.xsd",
    "http://www.xbrl.org/us/fr/rpt/mr/2005-02-28": "http://www.xbrl.org/us/fr/rpt/mr/2005-02-28/usfr-mr-2005-02-28.xsd",
    "http://www.xbrl.org/us/fr/rpt/seccert/2005-02-28": "http://www.xbrl.org/us/fr/rpt/seccert/2005-02-28/usfr-seccert-2005-02-28.xsd",
    "http://xbrl.ifrs.org/taxonomy/2013-09-09/ifrs": "http://xbrl.ifrs.org/taxonomy/2013-09-09/ifrs-cor_2013-09-09.xsd",
    "http://xbrl.ifrs.org/taxonomy/2014-03-05/ifrs-full": "http://xbrl.ifrs.org/taxonomy/2014-03-05/full_ifrs/full_ifrs-cor_2014-03-05.xsd",
    "http://xbrl.ifrs.org/taxonomy/2014-03-05/ifrs-smes": "http://xbrl.ifrs.org/taxonomy/2014-03-05/ifrs_for_smes/ifrs_for_smes-cor_2014-03-05.xsd",
    "http://xbrl.ifrs.org/taxonomy/2015-03-11/ifrs-full": "http://xbrl.ifrs.org/taxonomy/2015-03-11/full_ifrs/full_ifrs-cor_2015-03-11.xsd",
    "http://xbrl.ifrs.org/taxonomy/2016-03-31/ifrs-full": "http://xbrl.ifrs.org/taxonomy/2016-03-31/full_ifrs/full_ifrs-cor_2016-03-31.xsd",
    "http://xbrl.ifrs.org/taxonomy/2017-03-09/ifrs-full": "http://xbrl.ifrs.org/taxonomy/2017-03-09/full_ifrs/full_ifrs-cor_2017-03-09.xsd",
    "http://xbrl.ifrs.org/taxonomy/2018-03-16/ifrs-full": "http://xbrl.ifrs.org/taxonomy/2018-03-16/full_ifrs/full_ifrs-cor_2018-03-16.xsd",
    "http://xbrl.ifrs.org/taxonomy/2019-03-27/ifrs-full": "http://xbrl.ifrs.org/taxonomy/2019-03-27/full_ifrs/full_ifrs-cor_2019-03-27.xsd",
    "http://xbrl.ifrs.org/taxonomy/2020-03-16/ifrs-full": "http://xbrl.ifrs.org/taxonomy/2020-03-16/full_ifrs/full_ifrs-cor_2020-03-16.xsd",
    "http://xbrl.ifrs.org/taxonomy/2021-03-24/ifrs-full": "http://xbrl.ifrs.org/taxonomy/2021-03-24/full_ifrs/full_ifrs-cor_2021-03-24.xsd",
    "http://xbrl.org/2005/xbrldt": "http://www.xbrl.org/2005/xbrldt-2005.xsd",
    "http://xbrl.org/2006/xbrldi": "http://www.xbrl.org/2006/xbrldi-2006.xsd",
    "http://xbrl.org/2023/calculation-1.1": "https://www.xbrl.org/2023/calculation-1.1.xsd",
    "http://xbrl.org/2014/extensible-enumerations": "http://www.xbrl.org/2014/extensible-enumerations.xsd",
    "http://xbrl.org/2020/extensible-enumerations-2.0": "http://www.xbrl.org/2020/extensible-enumerations-2.0.xsd",
    "http://xbrl.sec.gov/country/2011-01-31": "https://xbrl.sec.gov/country/2011/country-2011-01-31.xsd",
    "http://xbrl.sec.gov/country/2012-01-31": "https://xbrl.sec.gov/country/2012/country-2012-01-31.xsd",
    "http://xbrl.sec.gov/country/2013-01-31": "https://xbrl.sec.gov/country/2013/country-2013-01-31.xsd",
    "http://xbrl.sec.gov/country/2016-01-31": "https://xbrl.sec.gov/country/2016/country-2016-01-31.xsd",
    "http://xbrl.sec.gov/country/2017-01-31": "https://xbrl.sec.gov/country/2017/country-2017-01-31.xsd",
    "http://xbrl.sec.gov/country/2020-01-31": "https://xbrl.sec.gov/country/2020/country-2020-01-31.xsd",
    "http://xbrl.sec.gov/country/2021": "https://xbrl.sec.gov/country/2021/country-2021.xsd",
    "http://xbrl.sec.gov/currency/2011-01-31": "https://xbrl.sec.gov/currency/2011/currency-2011-01-31.xsd",
    "http://xbrl.sec.gov/currency/2012-01-31": "https://xbrl.sec.gov/currency/2012/currency-2012-01-31.xsd",
    "http://xbrl.sec.gov/currency/2014-01-31": "https://xbrl.sec.gov/currency/2014/currency-2014-01-31.xsd",
    "http://xbrl.sec.gov/currency/2016-01-31": "https://xbrl.sec.gov/currency/2016/currency-2016-01-31.xsd",
    "http://xbrl.sec.gov/currency/2017-01-31": "https://xbrl.sec.gov/currency/2017/currency-2017-01-31.xsd",
    "http://xbrl.sec.gov/currency/2019-01-31": "https://xbrl.sec.gov/currency/2019/currency-2019-01-31.xsd",
    "http://xbrl.sec.gov/currency/2020-01-31": "https://xbrl.sec.gov/currency/2020/currency-2020-01-31.xsd",
    "http://xbrl.sec.gov/currency/2021": "https://xbrl.sec.gov/currency/2021/currency-2021.xsd",
    "http://xbrl.sec.gov/currency/2022": "https://xbrl.sec.gov/currency/2022/currency-2022.xsd",
    "http://xbrl.sec.gov/dei/2011-01-31": "https://xbrl.sec.gov/dei/2011/dei-2011-01-31.xsd",
    "http://xbrl.sec.gov/dei/2012-01-31": "https://xbrl.sec.gov/dei/2012/dei-2012-01-31.xsd",
    "http://xbrl.sec.gov/dei/2013-01-31": "https://xbrl.sec.gov/dei/2013/dei-2013-01-31.xsd",
    "http://xbrl.sec.gov/dei/2014-01-31": "https://xbrl.sec.gov/dei/2014/dei-2014-01-31.xsd",
    "http://xbrl.sec.gov/dei/2018-01-31": "https://xbrl.sec.gov/dei/2018/dei-2018-01-31.xsd",
    "http://xbrl.sec.gov/dei/2019-01-31": "https://xbrl.sec.gov/dei/2019/dei-2019-01-31.xsd",
    "http://xbrl.sec.gov/dei/2020-01-31": "https://xbrl.sec.gov/dei/2020/dei-2020-01-31.xsd",
    "http://xbrl.sec.gov/dei/2021": "https://xbrl.sec.gov/dei/2021/dei-2021.xsd",
    "http://xbrl.sec.gov/dei/2022": "https://xbrl.sec.gov/dei/2022/dei-2022.xsd",
    "http://xbrl.sec.gov/dei/2023": "https://xbrl.sec.gov/dei/2023/dei-2023.xsd",
    "http://xbrl.sec.gov/dei/2024": "https://xbrl.sec.gov/dei/2024/dei-2024.xsd",
    "http://xbrl.sec.gov/dei/2021q4": "https://xbrl.sec.gov/dei/2021q4/dei-2021q4.xsd",
    "http://xbrl.sec.gov/dei/2025": "https://xbrl.sec.gov/dei/2025/dei-2025.xsd",
    "http://xbrl.sec.gov/dei-def/2021": "https://xbrl.sec.gov/dei/2021/dei-2021_def.xsd",
    "http://xbrl.sec.gov/dei-entire/2021": "https://xbrl.sec.gov/dei/2021/dei-entire-2021.xsd",
    "http://xbrl.sec.gov/dei-ent-std/2019-01-31": "https://xbrl.sec.gov/dei/2019/dei-ent-std-2019-01-31.xsd",
    "http://xbrl.sec.gov/dei-ent-std/2020-01-31": "https://xbrl.sec.gov/dei/2020/dei-ent-std-2020-01-31.xsd",
    "http://xbrl.sec.gov/dei-lab/2021": "https://xbrl.sec.gov/dei/2021/dei-2021_lab.xsd",
    "http://xbrl.sec.gov/dei-pre/2021": "https://xbrl.sec.gov/dei/2021/dei-2021_pre.xsd",
    "http://xbrl.sec.gov/dei-std/2019-01-31": "https://xbrl.sec.gov/dei/2019/dei-std-2019-01-31.xsd",
    "http://xbrl.sec.gov/dei-std/2020-01-31": "https://xbrl.sec.gov/dei/2020/dei-std-2020-01-31.xsd",
    "http://xbrl.sec.gov/exch/2011-01-31": "https://xbrl.sec.gov/exch/2011/exch-2011-01-31.xsd",
    "http://xbrl.sec.gov/exch/2012-01-31": "https://xbrl.sec.gov/exch/2012/exch-2012-01-31.xsd",
    "http://xbrl.sec.gov/exch/2013-01-31": "https://xbrl.sec.gov/exch/2013/exch-2013-01-31.xsd",
    "http://xbrl.sec.gov/exch/2014-01-31": "https://xbrl.sec.gov/exch/2014/exch-2014-01-31.xsd",
    "http://xbrl.sec.gov/exch/2015-01-31": "https://xbrl.sec.gov/exch/2015/exch-2015-01-31.xsd",
    "http://xbrl.sec.gov/exch/2016-01-31": "https://xbrl.sec.gov/exch/2016/exch-2016-01-31.xsd",
    "http://xbrl.sec.gov/exch/2017-01-31": "https://xbrl.sec.gov/exch/2017/exch-2017-01-31.xsd",
    "http://xbrl.sec.gov/exch/2018-01-31": "https://xbrl.sec.gov/exch/2018/exch-2018-01-31.xsd",
    "http://xbrl.sec.gov/exch/2019-01-31": "https://xbrl.sec.gov/exch/2019/exch-2019-01-31.xsd",
    "http://xbrl.sec.gov/exch/2020-01-31": "https://xbrl.sec.gov/exch/2020/exch-2020-01-31.xsd",
    "http://xbrl.sec.gov/exch/2021": "https://xbrl.sec.gov/exch/2021/exch-2021.xsd",
    "http://xbrl.sec.gov/exch/2022": "https://xbrl.sec.gov/exch/2022/exch-2022.xsd",
    "http://xbrl.sec.gov/exch/2023": "https://xbrl.sec.gov/exch/2023/exch-2023.xsd",
    "http://xbrl.sec.gov/exch-def/2021": "https://xbrl.sec.gov/exch/2021/exch-2021_def.xsd",
    "http://xbrl.sec.gov/exch-lab/2021": "https://xbrl.sec.gov/exch/2021/exch-2021_lab.xsd",
    "http://xbrl.sec.gov/exch-pre/2021": "https://xbrl.sec.gov/exch/2021/exch-2021_pre.xsd",
    "http://xbrl.sec.gov/invest/2011-01-31": "https://xbrl.sec.gov/invest/2011/invest-2011-01-31.xsd",
    "http://xbrl.sec.gov/invest/2012-01-31": "https://xbrl.sec.gov/invest/2012/invest-2012-01-31.xsd",
    "http://xbrl.sec.gov/invest/2013-01-31": "https://xbrl.sec.gov/invest/2013/invest-2013-01-31.xsd",
    "http://xbrl.sec.gov/naics/2011-01-31": "https://xbrl.sec.gov/naics/2011/naics-2011-01-31.xsd",
    "http://xbrl.sec.gov/naics/2017-01-31": "https://xbrl.sec.gov/naics/2017/naics-2017-01-31.xsd",
    "http://xbrl.sec.gov/naics/2021": "https://xbrl.sec.gov/naics/2021/naics-2021.xsd",
    "http://xbrl.sec.gov/naics/2022": "https://xbrl.sec.gov/naics/2022/naics-2022.xsd",
    "http://xbrl.sec.gov/naics/2023": "https://xbrl.sec.gov/naics/2023/naics-2023.xsd",
    "http://xbrl.sec.gov/rr/2010-02-28": "https://xbrl.sec.gov/rr/2010/rr-2010-02-28.xsd",
    "http://xbrl.sec.gov/rr/2012-01-31": "https://xbrl.sec.gov/rr/2012/rr-2012-01-31.xsd",
    "http://xbrl.sec.gov/rr/2018-01-31": "https://xbrl.sec.gov/rr/2018/rr-2018-01-31.xsd",
    "http://xbrl.sec.gov/rr/2021": "https://xbrl.sec.gov/rr/2021/rr-2021.xsd",
    "http://xbrl.sec.gov/rr-cal/2010-02-28": "https://xbrl.sec.gov/rr/2010/rr-cal-2010-02-28.xsd",
    "http://xbrl.sec.gov/rr-cal/2012-01-31": "https://xbrl.sec.gov/rr/2012/rr-cal-2012-01-31.xsd",
    "http://xbrl.sec.gov/rr-cal/2018-01-31": "https://xbrl.sec.gov/rr/2018/rr-cal-2018-01-31.xsd",
    "http://xbrl.sec.gov/rr-def/2010-02-28": "https://xbrl.sec.gov/rr/2010/rr-def-2010-02-28.xsd",
    "http://xbrl.sec.gov/rr-def/2012-01-31": "https://xbrl.sec.gov/rr/2012/rr-def-2012-01-31.xsd",
    "http://xbrl.sec.gov/rr-def/2018-01-31": "https://xbrl.sec.gov/rr/2018/rr-def-2018-01-31.xsd",
    "http://xbrl.sec.gov/rr-def/2021": "https://xbrl.sec.gov/rr/2021/rr-2021_def.xsd",
    "http://xbrl.sec.gov/rr-ent/2010-02-28": "https://xbrl.sec.gov/rr/2010/rr-ent-2010-02-28.xsd",
    "http://xbrl.sec.gov/rr-ent/2012-01-31": "https://xbrl.sec.gov/rr/2012/rr-ent-2012-01-31.xsd",
    "http://xbrl.sec.gov/rr-ent/2018-01-31": "https://xbrl.sec.gov/rr/2018/rr-ent-2018-01-31.xsd",
    "http://xbrl.sec.gov/rr-lab/2021": "https://xbrl.sec.gov/rr/2021/rr-2021_lab.xsd",
    "http://xbrl.sec.gov/rr-pre/2010-02-28": "https://xbrl.sec.gov/rr/2010/rr-pre-2010-02-28.xsd",
    "http://xbrl.sec.gov/rr-pre/2012-01-31": "https://xbrl.sec.gov/rr/2012/rr-pre-2012-01-31.xsd",
    "http://xbrl.sec.gov/rr-pre/2018-01-31": "https://xbrl.sec.gov/rr/2018/rr-pre-2018-01-31.xsd",
    "http://xbrl.sec.gov/rr-pre/2021": "https://xbrl.sec.gov/rr/2021/rr-2021_pre.xsd",
    "http://xbrl.sec.gov/sic/2011-01-31": "https://xbrl.sec.gov/sic/2011/sic-2011-01-31.xsd",
    "http://xbrl.sec.gov/sic/2020-01-31": "https://xbrl.sec.gov/sic/2020/sic-2020-01-31.xsd",
    "http://xbrl.sec.gov/sic/2021": "https://xbrl.sec.gov/sic/2021/sic-2021.xsd",
    "http://xbrl.sec.gov/stpr/2011-01-31": "https://xbrl.sec.gov/stpr/2011/stpr-2011-01-31.xsd",
    "http://xbrl.sec.gov/stpr/2018-01-31": "https://xbrl.sec.gov/stpr/2018/stpr-2018-01-31.xsd",
    "http://xbrl.sec.gov/stpr/2021": "https://xbrl.sec.gov/stpr/2021/stpr-2021.xsd",
    "http://xbrl.sec.gov/stpr/2022": "https://xbrl.sec.gov/stpr/2022/stpr-2022.xsd",
    "http://xbrl.us/ar/2008-03-31": "http://xbrl.us/us-gaap/1.0/non-gaap/ar-2008-03-31.xsd",
    "http://xbrl.us/ar/2009-01-31": "http://taxonomies.xbrl.us/us-gaap/2009/non-gaap/ar-2009-01-31.xsd",
    "http://xbrl.us/country/2008-03-31": "http://xbrl.us/us-gaap/1.0/non-gaap/country-2008-03-31.xsd",
    "http://xbrl.us/country/2009-01-31": "http://taxonomies.xbrl.us/us-gaap/2009/non-gaap/country-2009-01-31.xsd",
    "http://xbrl.us/currency/2008-03-31": "http://xbrl.us/us-gaap/1.0/non-gaap/currency-2008-03-31.xsd",
    "http://xbrl.us/currency/2009-01-31": "http://taxonomies.xbrl.us/us-gaap/2009/non-gaap/currency-2009-01-31.xsd",
    "http://xbrl.us/dei/2008-03-31": "http://xbrl.us/us-gaap/1.0/non-gaap/dei-2008-03-31.xsd",
    "http://xbrl.us/dei/2009-01-31": "http://taxonomies.xbrl.us/us-gaap/2009/non-gaap/dei-2009-01-31.xsd",
    "http://xbrl.us/dei-ent/2008-03-31": "http://xbrl.us/us-gaap/1.0/non-gaap/dei-ent-2008-03-31.xsd",
    "http://xbrl.us/dei-std/2008-03-31": "http://xbrl.us/us-gaap/1.0/non-gaap/dei-std-2008-03-31.xsd",
    "http://xbrl.us/dei-std/2009-01-31": "http://taxonomies.xbrl.us/us-gaap/2009/non-gaap/dei-std-2009-01-31.xsd",
    "http://xbrl.us/exch/2008-03-31": "http://xbrl.us/us-gaap/1.0/non-gaap/exch-2008-03-31.xsd",
    "http://xbrl.us/exch/2009-01-31": "http://taxonomies.xbrl.us/us-gaap/2009/non-gaap/exch-2009-01-31.xsd",
    "http://xbrl.us/invest/2009-01-31": "http://taxonomies.xbrl.us/us-gaap/2009/non-gaap/invest-2009-01-31.xsd",
    "http://xbrl.us/mda/2008-03-31": "http://xbrl.us/us-gaap/1.0/non-gaap/mda-2008-03-31.xsd",
    "http://xbrl.us/mda/2009-01-31": "http://taxonomies.xbrl.us/us-gaap/2009/non-gaap/mda-2009-01-31.xsd",
    "http://xbrl.us/mr/2008-03-31": "http://xbrl.us/us-gaap/1.0/non-gaap/mr-2008-03-31.xsd",
    "http://xbrl.us/mr/2009-01-31": "http://taxonomies.xbrl.us/us-gaap/2009/non-gaap/mr-2009-01-31.xsd",
    "http://xbrl.us/naics/2008-03-31": "http://xbrl.us/us-gaap/1.0/non-gaap/naics-2008-03-31.xsd",
    "http://xbrl.us/naics/2009-01-31": "http://taxonomies.xbrl.us/us-gaap/2009/non-gaap/naics-2009-01-31.xsd",
    "http://xbrl.us/rr/2008-12-31": "http://taxonomies.xbrl.us/rr/2008/rr-2008-12-31.xsd",
    "http://xbrl.us/rr-ent/2008-12-31": "http://taxonomies.xbrl.us/rr/2008/rr-ent-2008-12-31.xsd",
    "http://xbrl.us/rr-std/2008-12-31": "http://taxonomies.xbrl.us/rr/2008/rr-std-2008-12-31.xsd",
    "http://xbrl.us/seccert/2008-03-31": "http://xbrl.us/us-gaap/1.0/non-gaap/seccert-2008-03-31.xsd",
    "http://xbrl.us/seccert/2009-01-31": "http://taxonomies.xbrl.us/us-gaap/2009/non-gaap/seccert-2009-01-31.xsd",
    "http://xbrl.us/sic/2008-03-31": "http://xbrl.us/us-gaap/1.0/non-gaap/sic-2008-03-31.xsd",
    "http://xbrl.us/sic/2009-01-31": "http://taxonomies.xbrl.us/us-gaap/2009/non-gaap/sic-2009-01-31.xsd",
    "http://xbrl.us/soi/2008-11-30": "http://taxonomies.xbrl.us/soi/2008/soi-2008-11-30.xsd",
    "http://xbrl.us/stpr/2008-03-31": "http://xbrl.us/us-gaap/1.0/non-gaap/stpr-2008-03-31.xsd",
    "http://xbrl.us/stpr/2009-01-31": "http://taxonomies.xbrl.us/us-gaap/2009/non-gaap/stpr-2009-01-31.xsd",
    "http://xbrl.us/us-gaap/2008-03-31": "http://xbrl.us/us-gaap/1.0/elts/us-gaap-2008-03-31.xsd",
    "http://xbrl.us/us-gaap/2009-01-31": "http://taxonomies.xbrl.us/us-gaap/2009/elts/us-gaap-2009-01-31.xsd",
    "http://xbrl.us/us-gaap/negated/2008-03-31": "http://www.xbrl.org/lrr/role/negated-2008-03-31.xsd",
    "http://xbrl.us/us-roles/2008-03-31": "http://xbrl.us/us-gaap/1.0/elts/us-roles-2008-03-31.xsd",
    "http://xbrl.us/us-roles/2009-01-31": "http://taxonomies.xbrl.us/us-gaap/2009/elts/us-roles-2009-01-31.xsd",
    "http://xbrl.us/us-types/2008-03-31": "http://xbrl.us/us-gaap/1.0/elts/us-types-2008-03-31.xsd",
    "http://xbrl.us/us-types/2009-01-31": "http://taxonomies.xbrl.us/us-gaap/2009/elts/us-types-2009-01-31.xsd",
    "http://xbrl.sec.gov/rxp/2023": "https://xbrl.sec.gov/rxp/2023/rxp-2023.xsd",
    "http://xbrl.sec.gov/snj/2023": "https://xbrl.sec.gov/snj/2023/snj-2023.xsd",
    "http://xbrl.sec.gov/snj-def/2023": "https://xbrl.sec.gov/snj/2023/snj-2023_def.xsd",
    "http://xbrl.sec.gov/country-def/2023": "https://xbrl.sec.gov/country/2023/country-2023_def.xsd",
    "http://xbrl.sec.gov/oef/2023": "https://xbrl.sec.gov/oef/2023/oef-2023.xsd",
    "http://xbrl.sec.gov/oef-rr/2023": "https://xbrl.sec.gov/oef/2023/oef-rr-2023.xsd",
    "http://xbrl.sec.gov/oef-sr/2023": "https://xbrl.sec.gov/oef/2023/oef-sr-2023.xsd",
    "http://xbrl.sec.gov/oef-lab/2023": "https://xbrl.sec.gov/oef/2023/oef-2023_lab.xsd",
    "http://xbrl.sec.gov/oef-cal/2023": "https://xbrl.sec.gov/oef/2023/oef-2023_cal.xsd",
    "https://xbrl.ifrs.org/taxonomy/2023-03-23/ifrs-full": "https://xbrl.ifrs.org/taxonomy/2023-03-23/full_ifrs/full_ifrs-cor_2023-03-23.xsd",
    "http://xbrl.sec.gov/rr-sub/2023": "https://xbrl.sec.gov/rr/2023/rr-sub-2023.xsd",
    "http://xbrl.sec.gov/rr-cal/2023": "https://xbrl.sec.gov/rr/2023/rr-2023_cal.xsd",
    "http://xbrl.sec.gov/dei-sub/2023": "https://xbrl.sec.gov/dei/2023/dei-sub-2023.xsd",
    "http://xbrl.sec.gov/dei-def/2023": "https://xbrl.sec.gov/dei/2023/dei-2023_def.xsd",
    "http://xbrl.sec.gov/dei-lab/2023": "https://xbrl.sec.gov/dei/2023/dei-2023_lab.xsd",
    "http://xbrl.sec.gov/dei-pre/2023": "https://xbrl.sec.gov/dei/2023/dei-2023_pre.xsd",
    "http://xbrl.sec.gov/dei/2022q4": "https://xbrl.sec.gov/dei/2022q4/dei-2022q4.xsd",
    "http://xbrl.sec.gov/dei-sub/2022q4": "https://xbrl.sec.gov/dei/2022q4/dei-sub-2022q4.xsd",
    "http://xbrl.sec.gov/dei-def/2022q4": "https://xbrl.sec.gov/dei/2022q4/dei-2022q4_def.xsd",
    "http://xbrl.sec.gov/dei-lab/2022q4": "https://xbrl.sec.gov/dei/2022q4/dei-2022q4_lab.xsd",
    "http://xbrl.sec.gov/dei-pre/2022q4": "https://xbrl.sec.gov/dei/2022q4/dei-2022q4_pre.xsd",
    "http://xbrl.sec.gov/ecd/2023": "https://xbrl.sec.gov/ecd/2023/ecd-2023.xsd",
    "http://xbrl.sec.gov/ecd-sub/2023": "https://xbrl.sec.gov/ecd/2023/ecd-sub-2023.xsd",
    "http://xbrl.sec.gov/ecd/2022q4": "https://xbrl.sec.gov/ecd/2022q4/ecd-2022q4.xsd",
    "http://xbrl.sec.gov/ecd-sub/2022q4": "https://xbrl.sec.gov/ecd/2022q4/ecd-sub-2022q4.xsd",
    "http://fasb.org/srt-sup/2022q3": "https://xbrl.fasb.org/srt/2022q3/srt-sup-2022q3.xsd",
    "http://fasb.org/us-gaap-sup/2022q3": "https://xbrl.fasb.org/us-gaap/2022q3/us-gaap-sup-2022q3.xsd",
    "http://xbrl.sec.gov/vip/2023": "https://xbrl.sec.gov/vip/2023/vip-2023.xsd",
    "http://xbrl.sec.gov/vip-n3/2023": "https://xbrl.sec.gov/vip/2023/vip-n3-2023.xsd",
    "http://xbrl.sec.gov/vip-n4/2023": "https://xbrl.sec.gov/vip/2023/vip-n4-2023.xsd",
    "http://xbrl.sec.gov/vip-n6/2023": "https://xbrl.sec.gov/vip/2023/vip-n6-2023.xsd",
    "http://xbrl.sec.gov/country/2023": "https://xbrl.sec.gov/country/2023/country-2023.xsd",
    "http://xbrl.sec.gov/currency/2023": "https://xbrl.sec.gov/currency/2023/currency-2023.xsd",
    "http://xbrl.sec.gov/vip/2022q2": "https://xbrl.sec.gov/vip/2022q2/vip-2022q2.xsd",
    "http://xbrl.sec.gov/vip-n3/2022q2": "https://xbrl.sec.gov/vip/2022q2/vip-n3-2022q2.xsd",
    "http://xbrl.sec.gov/vip-n4/2022q2": "https://xbrl.sec.gov/vip/2022q2/vip-n4-2022q2.xsd",
    "http://xbrl.sec.gov/vip-n6/2022q2": "https://xbrl.sec.gov/vip/2022q2/vip-n6-2022q2.xsd",
    "http://xbrl.sec.gov/country/2022": "https://xbrl.sec.gov/country/2022/country-2022.xsd",
    "http://xbrl.sec.gov/sic/2022": "https://xbrl.sec.gov/sic/2022/sic-2022.xsd",
    "https://xbrl.ifrs.org/taxonomy/2022-03-24/ifrs-full": "https://xbrl.ifrs.org/taxonomy/2022-03-24/full_ifrs/full_ifrs-cor_2022-03-24.xsd",
    "http://xbrl.sec.gov/dei-lab/2022": "https://xbrl.sec.gov/dei/2022/dei-2022_lab.xsd",
    "http://xbrl.sec.gov/dei-pre/2022": "https://xbrl.sec.gov/dei/2022/dei-2022_pre.xsd",
    "http://xbrl.sec.gov/dei-def/2022": "https://xbrl.sec.gov/dei/2022/dei-2022_def.xsd",
    "http://xbrl.sec.gov/sic/2023": "https://xbrl.sec.gov/sic/2023/sic-2023.xsd",
    "http://xbrl.sec.gov/stpr/2023": "https://xbrl.sec.gov/stpr/2023/stpr-2023.xsd",
    "http://fasb.org/us-types/2023": "https://xbrl.fasb.org/us-gaap/2023/elts/us-types-2023.xsd",
    "http://fasb.org/us-roles/2023": "https://xbrl.fasb.org/us-gaap/2023/elts/us-roles-2023.xsd",
    "http://fasb.org/srt-types/2023": "https://xbrl.fasb.org/srt/2023/elts/srt-types-2023.xsd",
    "http://fasb.org/srt-roles/2023": "https://xbrl.fasb.org/srt/2023/elts/srt-roles-2023.xsd",
    "http://xbrl.sec.gov/dei-sub/2022": "https://xbrl.sec.gov/dei/2022/dei-sub-2022.xsd",
    "http://fasb.org/us-types/2022": "https://xbrl.fasb.org/us-gaap/2022/elts/us-types-2022.xsd",
    "http://fasb.org/us-roles/2022": "https://xbrl.fasb.org/us-gaap/2022/elts/us-roles-2022.xsd",
    "http://fasb.org/srt-types/2022": "https://xbrl.fasb.org/srt/2022/elts/srt-types-2022.xsd",
    "http://fasb.org/srt-roles/2022": "https://xbrl.fasb.org/srt/2022/elts/srt-roles-2022.xsd",
    "http://xbrl.sec.gov/cef/2023": "https://xbrl.sec.gov/cef/2023/cef-2023.xsd",
    "http://xbrl.sec.gov/cef-pre/2023": "https://xbrl.sec.gov/cef/2023/cef-2023_pre.xsd",
    "http://xbrl.sec.gov/cef/2022": "https://xbrl.sec.gov/cef/2022/cef-2022.xsd",
    "http://xbrl.sec.gov/vip/2022": "https://xbrl.sec.gov/vip/2022/vip-2022.xsd",
    "http://xbrl.sec.gov/vip-n3/2022": "https://xbrl.sec.gov/vip/2022/vip-n3-2022.xsd",
    "http://xbrl.sec.gov/vip-n4/2022": "https://xbrl.sec.gov/vip/2022/vip-n4-2022.xsd",
    "http://xbrl.sec.gov/vip-n6/2022": "https://xbrl.sec.gov/vip/2022/vip-n6-2022.xsd",
    "http://xbrl.sec.gov/rr/2023": "https://xbrl.sec.gov/rr/2023/rr-2023.xsd",
    "http://xbrl.sec.gov/rr-lab/2023": "https://xbrl.sec.gov/rr/2023/rr-2023_lab.xsd",
    "http://xbrl.sec.gov/rr-pre/2023": "https://xbrl.sec.gov/rr/2023/rr-2023_pre.xsd",
    "http://xbrl.sec.gov/rr-def/2023": "https://xbrl.sec.gov/rr/2023/rr-2023_def.xsd",
    "http://xbrl.sec.gov/rr/2022": "https://xbrl.sec.gov/rr/2022/rr-2022.xsd",
    "http://xbrl.sec.gov/rr-lab/2022": "https://xbrl.sec.gov/rr/2022/rr-2022_lab.xsd",
    "http://xbrl.sec.gov/rr-pre/2022": "https://xbrl.sec.gov/rr/2022/rr-2022_pre.xsd",
    "http://xbrl.sec.gov/rr-def/2022": "https://xbrl.sec.gov/rr/2022/rr-2022_def.xsd",
    "http://xbrl.sec.gov/cyd/2024": "https://xbrl.sec.gov/cyd/2024/cyd-2024.xsd",
    "http://xbrl.sec.gov/cyd-entire/2024": "https://xbrl.sec.gov/cyd/2024/cyd-entire-2024.xsd",
    "http://xbrl.sec.gov/cyd/2025": "https://xbrl.sec.gov/cyd/2025/cyd-2025.xsd",
    "http://xbrl.sec.gov/cyd-af/2025": "https://xbrl.sec.gov/cyd/2025/cyd-af-2025.xsd",
    "http://xbrl.sec.gov/cyd-af-sub/2025": "https://xbrl.sec.gov/cyd/2025/cyd-af-sub-2025.xsd",
    "http://www.xbrl.org/dtr/type/2022-03-31": "https://www.xbrl.org/dtr/type/2022-03-31/types.xsd",
    "http://www.hmrc.gov.uk/schemas/ct/dpl/2021-01-01": "https://www.hmrc.gov.uk/schemas/ct/dpl/2021-01-01/dpl-2021.xsd",
    "http://xbrl.sec.gov/ffd/2025": "https://xbrl.sec.gov/ffd/2025/ffd-2025.xsd",
}
